import datetime
import logging
import resource
import sys
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Sequence, Union

from seml.database import get_collection
from seml.experiment.observers import create_mongodb_observer
from seml.settings import SETTINGS
from seml.utils.multi_process import is_main_process

# These are only used for type hints
if TYPE_CHECKING:
    from sacred import Experiment as ExperimentBase
    from sacred import Ingredient
    from sacred.commandline_options import CLIOption
    from sacred.host_info import HostInfoGetter
    from sacred.utils import PathType
from sacred import SETTINGS as SACRED_SETTINGS
from sacred import Experiment as ExperimentBase
from sacred.config.utils import (
    dogmatize,
    recursive_fill_in,
    undogmatize,
)


class LoggerOptions(Enum):
    NONE = None
    DEFAULT = 'default'
    RICH = 'rich'


class Experiment(ExperimentBase):
    def __init__(
        self,
        name: Optional[str] = None,
        ingredients: Sequence['Ingredient'] = (),
        interactive: bool = False,
        base_dir: Optional['PathType'] = None,
        additional_host_info: Optional[List['HostInfoGetter']] = None,
        additional_cli_options: Optional[Sequence['CLIOption']] = None,
        save_git_info: bool = True,
        add_mongodb_observer: bool = True,
        logger: Optional[Union[LoggerOptions, str]] = LoggerOptions.RICH,
        capture_output: Optional[bool] = None,
        collect_stats: bool = True,
    ):
        super().__init__(
            name=name,
            ingredients=ingredients,
            interactive=interactive,
            base_dir=base_dir,
            additional_host_info=additional_host_info,
            additional_cli_options=additional_cli_options,
            save_git_info=save_git_info,
        )
        self.capture_output = capture_output
        if add_mongodb_observer:
            self.configurations.append(MongoDbObserverConfig(self))
        self.configurations.append(ClearObserverForMultiTaskConfig(self))
        if logger:
            setup_logger(self, LoggerOptions(logger))
        if collect_stats:
            self.post_run_hook(lambda _run: _collect_exp_stats(_run))

    def run(
        self,
        command_name: Optional[str] = None,
        config_updates: Optional[dict] = None,
        named_configs: Sequence[str] = (),
        info: Optional[dict] = None,
        meta_info: Optional[dict] = None,
        options: Optional[dict] = None,
    ):
        if (
            not SETTINGS.EXPERIMENT.CAPTURE_OUTPUT and not self.capture_output
        ) or self.capture_output is False:
            SACRED_SETTINGS.CAPTURE_MODE = 'no'
        super().run(
            command_name=command_name,
            config_updates=config_updates,
            named_configs=named_configs,
            info=info,
            meta_info=meta_info,
            options=options,
        )


class MongoDbObserverConfig:
    def __init__(self, experiment: Experiment):
        self.experiment = experiment

    def __call__(self, fixed=None, preset=None, fallback=None):
        from sacred.config.config_summary import ConfigSummary

        result = dogmatize(fixed or {})
        defaults = dict(overwrite=None, db_collection=None)
        recursive_fill_in(result, defaults)
        recursive_fill_in(result, preset or {})
        added = result.revelation()
        config_summary = ConfigSummary(added, result.modified, result.typechanges)
        config_summary.update(undogmatize(result))

        if config_summary['db_collection'] is not None and is_main_process():
            self.experiment.observers.append(
                create_mongodb_observer(
                    config_summary['db_collection'],
                    overwrite=config_summary['overwrite'],
                )
            )
        return config_summary


class ClearObserverForMultiTaskConfig:
    def __init__(self, experiment: Experiment):
        self.experiment = experiment

    def __call__(self, fixed=None, preset=None, fallback=None):
        from sacred.config.config_summary import ConfigSummary

        result = dogmatize(fixed or {})
        defaults = dict(overwrite=None, db_collection=None)
        recursive_fill_in(result, defaults)
        recursive_fill_in(result, preset or {})
        added = result.revelation()
        config_summary = ConfigSummary(added, result.modified, result.typechanges)
        config_summary.update(undogmatize(result))

        # We only want observers on the main process
        if not is_main_process():
            self.experiment.observers.clear()
        return config_summary


def setup_logger(
    ex: 'ExperimentBase',
    logger_option: LoggerOptions = LoggerOptions.RICH,
    level: Optional[Union[str, int]] = None,
):
    """
    Set up logger for experiment.

    Parameters
    ----------
    ex: sacred.Experiment
    Sacred experiment to set the logger of.
    level: str or int
    Set the threshold for the logger to this level. Default is logging.INFO.

    Returns
    -------
    None

    """
    if hasattr(ex, 'logger') and ex.logger:
        logging.warn(
            'Logger already set up for this experiment.\n'
            'The new seml.experiment.Experiment class already includes the logger setup.\n'
            'Either remove the explicit call to setup_logger or disable the logger setup in the Experiment constructor.'
        )
        return
    if logger_option is LoggerOptions.NONE:
        return
    logger = logging.getLogger()
    logger.handlers = []
    if level is None:
        if is_main_process():
            level = logging.INFO
        else:
            level = logging.ERROR
    if logger_option is LoggerOptions.RICH:
        from rich.logging import RichHandler

        from seml.console import console

        logger.addHandler(
            RichHandler(
                level,
                console=console,
                show_time=True,
                show_level=True,
                log_time_format='[%X]',
            )
        )
    elif logger_option is LoggerOptions.DEFAULT:
        ch = logging.StreamHandler()
        formatter = logging.Formatter(
            fmt='%(asctime)s (%(levelname)s): %(message)s', datefmt='%Y-%m-%d %H:%M:%S'
        )
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    logger.setLevel(level)
    ex.logger = logger


def _collect_exp_stats(run):
    """
    Collect information such as CPU user time, maximum memory usage,
    and maximum GPU memory usage and save it in the MongoDB.

    Parameters
    ----------
    run: Sacred run
        Current Sacred run.

    Returns
    -------
    None
    """
    exp_id = run.config['overwrite']
    if exp_id is None or run.unobserved:
        return

    stats = {}

    stats['real_time'] = (datetime.datetime.utcnow() - run.start_time).total_seconds()

    stats['self'] = {}
    stats['self']['user_time'] = resource.getrusage(resource.RUSAGE_SELF).ru_utime
    stats['self']['system_time'] = resource.getrusage(resource.RUSAGE_SELF).ru_stime
    stats['self']['max_memory_bytes'] = (
        1024 * resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    )
    stats['children'] = {}
    stats['children']['user_time'] = resource.getrusage(
        resource.RUSAGE_CHILDREN
    ).ru_utime
    stats['children']['system_time'] = resource.getrusage(
        resource.RUSAGE_CHILDREN
    ).ru_stime
    stats['children']['max_memory_bytes'] = (
        1024 * resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
    )

    if 'torch' in sys.modules:
        import torch

        stats['pytorch'] = {}
        if torch.cuda.is_available():
            stats['pytorch']['gpu_max_memory_bytes'] = torch.cuda.max_memory_allocated()

    if 'tensorflow' in sys.modules:
        import tensorflow as tf

        stats['tensorflow'] = {}
        if int(tf.__version__.split('.')[0]) < 2:
            if tf.test.is_gpu_available():
                with tf.Session() as sess:
                    stats['tensorflow']['gpu_max_memory_bytes'] = int(
                        sess.run(tf.contrib.memory_stats.MaxBytesInUse())
                    )
        else:
            if len(tf.config.experimental.list_physical_devices('GPU')) >= 1:
                if int(tf.__version__.split('.')[1]) >= 5:
                    stats['tensorflow']['gpu_max_memory_bytes'] = (
                        tf.config.experimental.get_memory_info('GPU:0')['peak']
                    )
                else:
                    logging.info(
                        'SEML stats: There is no way to get actual peak GPU memory usage in TensorFlow 2.0-2.4.'
                    )

    collection = get_collection(run.config['db_collection'])
    collection.update_one({'_id': exp_id}, {'$set': {'stats': stats}})


def collect_exp_stats(run):
    logging.warn(
        'seml.collect_exp_stats is deprecated.\n'
        'Use seml.experiment.Experiment instead of sacred.Experiment.\n'
        'seml.experiment.Experiment already includes the statistics collection.\n'
        'See https://github.com/TUM-DAML/seml/blob/master/examples/example_experiment.py'
    )
    _collect_exp_stats(run)


__all__ = [
    'setup_logger',
    'collect_exp_stats',
    'Experiment',
]



#!/usr/bin/env python
import functools
import logging
import os
import sys
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, Set, Tuple, TypeVar

from typing_extensions import Annotated, ParamSpec

import seml.utils.typer as typer
from seml.commands.add import add_config_files
from seml.commands.configure import configure
from seml.commands.description import (
    collection_delete_description,
    collection_list_descriptions,
    collection_set_description,
)
from seml.commands.manage import (
    cancel_empty_pending_jobs,
    cancel_experiments,
    delete_experiments,
    detect_killed,
    drop_collections,
    reload_sources,
    reset_experiments,
)
from seml.commands.migration import migrate_collection
from seml.commands.print import (
    print_collections,
    print_command,
    print_duplicates,
    print_experiment,
    print_fail_trace,
    print_output,
    print_queue,
    print_status,
)
from seml.commands.project import init_project, print_available_templates
from seml.commands.slurm import hold_or_release_experiments
from seml.commands.sources import download_sources
from seml.commands.start import (
    claim_experiment,
    prepare_experiment,
    start_experiments,
    start_jupyter_job,
)
from seml.database import (
    clean_unreferenced_artifacts,
    get_collections_from_mongo_shell_or_pymongo,
    get_mongodb_config,
    update_working_dir,
)
from seml.settings import SETTINGS
from seml.utils import cache_to_disk
from seml.utils.module_hider import AUTOCOMPLETING

States = SETTINGS.STATES


P = ParamSpec('P')
R = TypeVar('R')


# numexpr will log unnecessary info we don't want in our CLI
logging.getLogger('numexpr').setLevel(logging.ERROR)

# Let's not import json if we are only autocompleting
if not AUTOCOMPLETING:
    import json

JsonOption = functools.partial(
    typer.Option,
    metavar='JSON',
    parser=json.loads if not AUTOCOMPLETING else lambda s: None,  # type: ignore
)


_EXPERIMENTS = '🚀 Experiments'
_DATABASE = '📊 Database'
_INFORMATION = '💭 Information'
_SLURM = '🏃 Slurm'


def restrict_collection(require: bool = True):
    """Decorator to require a collection name."""

    def decorator(fun: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(fun)
        def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs):
            if require and not ctx.obj['collection']:
                raise typer.BadParameter('Please specify a collection name.', ctx=ctx)
            elif not require and ctx.obj['collection']:
                raise typer.BadParameter(
                    'Please do not specify a collection name.', ctx=ctx
                )
            return fun(ctx, *args, **kwargs)  # type: ignore

        wrapper._requires_collection = require  # type: ignore
        return wrapper  # type: ignore

    return decorator


def collection_free_commands(app: typer.Typer) -> List[str]:
    """Get the commands that do not require a collection."""
    return [
        cmd.name if cmd.name else cmd.callback.__name__  # type: ignore
        for cmd in app.registered_commands
        if not getattr(cmd.callback, '_requires_collection', True)
    ]


@cache_to_disk('db_config', SETTINGS.AUTOCOMPLETE_CACHE_ALIVE_TIME)
def get_db_collections():
    """CLI completion for db collections."""
    config = get_mongodb_config()
    return list(get_collections_from_mongo_shell_or_pymongo(**config))


def first_argument_completer():
    """CLI completition for the first argumentin SEML."""
    # We also add the commands that do not require a collection for autocompletion.
    return get_db_collections() + collection_free_commands(app)


app = typer.Typer(
    no_args_is_help=True,
    # Note that this is not 100% the correct chaining autocompletition
    # but it is significantly better than nothing. Compared to the default
    # click chaining we greedly split the arguments by any command.
    chain=AUTOCOMPLETING,
)
YesAnnotation = Annotated[
    bool,
    typer.Option(
        '-y',
        '--yes',
        help='Automatically confirm all dialogues with yes.',
        is_flag=True,
    ),
]
SacredIdAnnotation = Annotated[
    Optional[int],
    typer.Option(
        '-id',
        '--sacred-id',
        help='Sacred ID (_id in the database collection) of the experiment. '
        'Takes precedence over other filters.',
    ),
]
FilterDictAnnotation = Annotated[
    Optional[Dict],
    JsonOption(
        '-f',
        '--filter-dict',
        help='Dictionary (passed as a string, e.g. \'{"config.dataset": "cora_ml"}\') to filter '
        'the experiments by.',
    ),
]
BatchIdAnnotation = Annotated[
    Optional[int],
    typer.Option(
        '-b',
        '--batch-id',
        help='Batch ID (batch_id in the database collection) of the experiments. '
        'Experiments that were staged together have the same batch_id.',
    ),
]


def parse_optional_str_list(values: Optional[Sequence[str]]) -> List[str]:
    if values is None:
        return []
    return [
        __x.strip()
        for _x in values
        for __x in _x.replace(',', ' ').split()
        if __x.strip()
    ]


ProjectionAnnotation = Annotated[
    List[str],
    typer.Option(
        '-p',
        '--projection',
        help='List of configuration keys, e.g., `config.model`, to additionally print.',
        parser=str.strip,
        callback=parse_optional_str_list,
        metavar='KEY',
    ),
]

_STATE_LIST = [s for states in States.values() for s in states]
FilterStatesAnnotation = Annotated[
    List[str],
    typer.Option(
        '-s',
        '--filter-states',
        help='List of states to filter the experiments by. If empty (""), all states are considered.',
        metavar=f'[{"|".join(_STATE_LIST)}]',
        parser=lambda s: s.strip().upper(),
        callback=parse_optional_str_list,
    ),
]
SBatchOptionsAnnotation = Annotated[
    Optional[Dict],
    JsonOption(
        '-sb',
        '--sbatch-options',
        help='Dictionary (passed as a string, e.g. \'{"gres": "gpu:2"}\') to request two GPUs.',
    ),
]
NumExperimentsAnnotation = Annotated[
    int,
    typer.Option(
        '-n',
        '--num-experiments',
        help='Number of experiments to start. ' '0: all (staged) experiments ',
    ),
]
NoFileOutputAnnotation = Annotated[
    bool,
    typer.Option(
        '-nf',
        '--no-file-output',
        help="Do not write the experiment's output to a file.",
        is_flag=True,
    ),
]
OutputToConsoleAnnotation = Annotated[
    bool,
    typer.Option(
        '-o',
        '--output-to-console',
        help="Write the experiment's output to the console.",
        is_flag=True,
    ),
]
StealSlurmAnnotation = Annotated[
    bool,
    typer.Option(
        '-ss',
        '--steal-slurm',
        help="Local jobs 'steal' from the Slurm queue, "
        'i.e. also execute experiments waiting for execution via Slurm.',
        is_flag=True,
    ),
]
PostMortemAnnotation = Annotated[
    bool,
    typer.Option(
        '-pm',
        '--post-mortem',
        help='Activate post-mortem debugging with pdb.',
        is_flag=True,
    ),
]
WorkerGPUsAnnotation = Annotated[
    Optional[str],
    typer.Option(
        '-wg',
        '--worker-gpus',
        help='The IDs of the GPUs used by the local worker. Will be directly passed to CUDA_VISIBLE_DEVICES.',
    ),
]
WorkerCPUsAnnotation = Annotated[
    Optional[int],
    typer.Option(
        '-wc',
        '--worker-cpus',
        help='The number of CPUs used by the local worker. Will be directly passed to OMP_NUM_THREADS.',
    ),
]
WorkerEnvAnnotation = Annotated[
    Optional[Dict],
    JsonOption(
        '-we',
        '--worker-env',
        help='Further environment variables to be set for the local worker.',
    ),
]
PrintFullDescriptionAnnotation = Annotated[
    bool,
    typer.Option(
        '-fd',
        '--full-descriptions',
        help='Whether to print full descriptions (possibly with line breaks).',
        is_flag=True,
    ),
]
UpdateStatusAnnotation = Annotated[
    bool,
    typer.Option(
        '-u',
        '--update-status',
        help='Whether to update the status of experiments in the database. '
        'This can take a while for large collections. Use only if necessary.',
        is_flag=True,
    ),
]
NoResolveDescriptionAnnotation = Annotated[
    bool,
    typer.Option(
        '--no-resolve-descriptions',
        help='Whether to prevent using omegaconf to resolve experiment descriptions',
        is_flag=True,
    ),
]
DebugAnnotation = Annotated[
    bool,
    typer.Option(
        '-d',
        '--debug',
        help='Run a single interactive experiment without Sacred observers and with post-mortem debugging. '
        'Implies `--verbose --num-exps 1 --post-mortem --output-to-console`.',
        is_flag=True,
    ),
]

DebugServerAnnotation = Annotated[
    bool,
    typer.Option(
        '-ds',
        '--debug-server',
        help='Run the experiment with a debug server, to which you can remotely connect with e.g. VS Code. '
        'Implies `--debug`.',
        is_flag=True,
    ),
]


def version_callback(value: bool):
    if value:
        from seml import __version__

        print(__version__)
        raise typer.Exit(0)


@app.callback()
def callback(
    ctx: typer.Context,
    collection: Annotated[
        str,
        typer.Argument(
            help='The name of the database collection to use.',
            autocompletion=first_argument_completer,
        ),
    ],
    migration_skip: Annotated[
        bool,
        typer.Option(
            '--migration-skip',
            help='Skip the migration of the database collection.',
            is_flag=True,
        ),
    ] = False,
    migration_backup: Annotated[
        bool,
        typer.Option(
            '--migration-backup',
            help='Backup the database collection before migration.',
            is_flag=True,
        ),
    ] = False,
    verbose: Annotated[
        bool,
        typer.Option(
            '-v',
            '--verbose',
            help='Whether to print debug messages.',
            is_flag=True,
        ),
    ] = False,
    version: Annotated[
        bool,
        typer.Option(
            '-V',
            '--version',
            help='Print the version number.',
            is_flag=True,
            callback=version_callback,
        ),
    ] = False,
):
    """SEML - Slurm Experiment Management Library."""
    from rich.logging import RichHandler

    from seml.console import console

    if len(logging.root.handlers) == 0:
        logging_level = logging.NOTSET if verbose else logging.INFO
        handler = RichHandler(
            logging_level,
            console=console,
            show_path=False,
            show_level=True,
            show_time=False,
        )
        logging.basicConfig(
            level=logging_level, format='%(message)s', handlers=[handler]
        )

    if collection:
        migrate_collection(collection, migration_skip, migration_backup)

    ctx.obj = dict(collection=collection, verbose=verbose)


@app.command('list', rich_help_panel=_INFORMATION)
@restrict_collection(False)
def list_command(
    ctx: typer.Context,
    pattern: Annotated[
        str, typer.Argument(help='A regex that must match the collections to print.')
    ] = r'.*',
    progress: Annotated[
        bool,
        typer.Option(
            '-p',
            '--progress',
            help='Whether to print a progress bar for iterating over collections.',
            is_flag=True,
        ),
    ] = False,
    update_status: UpdateStatusAnnotation = False,
    full_description: PrintFullDescriptionAnnotation = False,
):
    """Lists all collections in the database."""
    print_collections(
        pattern,
        progress=progress,
        update_status=update_status,
        print_full_description=full_description,
    )


@app.command('clean-db', rich_help_panel=_DATABASE)
def clean_db_command(ctx: typer.Context, yes: YesAnnotation = False):
    """Remove orphaned artifacts in the DB from runs which have been deleted.."""
    clean_unreferenced_artifacts(ctx.obj['collection'], yes=yes)


@app.command('configure')
@restrict_collection(False)
def configure_command(
    ctx: typer.Context,
    ssh_forward: Annotated[
        bool,
        typer.Option(
            '-sf',
            '--ssh-forward',
            help='Configure SSH forwarding settings for MongoDB.',
            is_flag=True,
        ),
    ] = False,
):
    """
    Configure SEML (database, argument completion, ...).
    """
    configure(all=False, mongodb=True, setup_ssh_forward=ssh_forward)


@app.command('start-jupyter', rich_help_panel=_SLURM)
@restrict_collection(False)
def start_jupyter_command(
    ctx: typer.Context,
    lab: Annotated[
        bool,
        typer.Option(
            '-l',
            '--lab',
            help='Start a jupyter-lab instance instead of jupyter notebook.',
        ),
    ] = False,
    conda_env: Annotated[
        Optional[str],
        typer.Option(
            '-c',
            '--conda-env',
            help='Start the Jupyter instance in a Conda environment.',
        ),
    ] = None,
    sbatch_options: SBatchOptionsAnnotation = None,
):
    """
    Start a Jupyter slurm job. Uses SBATCH options defined in settings.py under
    SBATCH_OPTIONS_TEMPLATES.JUPYTER
    """
    start_jupyter_job(lab=lab, conda_env=conda_env, sbatch_options=sbatch_options)


@app.command('cancel', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def cancel_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [*States.PENDING, *States.RUNNING],
    wait: Annotated[
        bool,
        typer.Option(
            '-w',
            '--wait',
            help='Wait until all jobs are properly cancelled.',
            is_flag=True,
        ),
    ] = False,
    yes: YesAnnotation = False,
):
    """
    Cancel the Slurm job/job step corresponding to experiments, filtered by ID or state.
    """
    wait |= (
        len(
            [
                a
                for a in sys.argv
                if a in command_tree(app).commands or a in command_tree(app).groups
            ]
        )
        > 1
    )
    cancel_experiments(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_dict=filter_dict,
        batch_id=batch_id,
        filter_states=filter_states,
        wait=wait,
        yes=yes,
    )


@app.command('add', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def add_command(
    ctx: typer.Context,
    config_files: Annotated[
        List[str],
        typer.Argument(
            help='Path to the YAML configuration file for the experiment.',
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
    ],
    no_hash: Annotated[
        bool,
        typer.Option(
            '-nh',
            '--no-hash',
            help='By default, we use the hash of the config dictionary to filter out duplicates (by comparing all '
            'dictionary values individually). Only disable this if you have a good reason as it is faster.',
            is_flag=True,
        ),
    ] = False,
    no_sanity_check: Annotated[
        bool,
        typer.Option(
            '-ncs',
            '--no-sanity-check',
            help='Disable this if the check fails unexpectedly when using '
            'advanced Sacred features or to accelerate adding.',
            is_flag=True,
        ),
    ] = False,
    no_code_checkpoint: Annotated[
        bool,
        typer.Option(
            '-ncc',
            '--no-code-checkpoint',
            help='Disable this if you want your experiments to use the current code'
            'instead of the code at the time of adding.',
            is_flag=True,
        ),
    ] = False,
    force: Annotated[
        bool,
        typer.Option(
            '-f',
            '--force',
            help='Force adding the experiment even if it already exists in the database.',
            is_flag=True,
        ),
    ] = False,
    overwrite_params: Annotated[
        Optional[Dict],
        JsonOption(
            '-o',
            '--overwrite-params',
            help='Dictionary (passed as a string, e.g. \'{"epochs": 100}\') to overwrite parameters in the config.',
        ),
    ] = None,
    description: Annotated[
        Optional[str],
        typer.Option(
            '-d',
            '--description',
            help='A description for the experiment.',
        ),
    ] = None,
    no_resolve_descriptions: NoResolveDescriptionAnnotation = False,
):
    """
    Add experiments to the database as defined in the configuration.
    """
    add_config_files(
        ctx.obj['collection'],
        config_files,
        force_duplicates=force,
        no_hash=no_hash,
        no_sanity_check=no_sanity_check,
        no_code_checkpoint=no_code_checkpoint,
        overwrite_params=overwrite_params,
        description=description,
        resolve_descriptions=not no_resolve_descriptions,
    )
    get_db_collections.recompute_cache()


@app.command('start', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def start_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    debug: DebugAnnotation = False,
    debug_server: DebugServerAnnotation = False,
    local: Annotated[
        bool,
        typer.Option(
            '-l',
            '--local',
            help='Run the experiment locally instead of on a Slurm cluster.',
            is_flag=True,
        ),
    ] = False,
    no_worker: Annotated[
        bool,
        typer.Option(
            '-nw',
            '--no-worker',
            help="Do not launch a local worker after setting experiments' state to PENDING.",
            is_flag=True,
        ),
    ] = False,
    num_exps: NumExperimentsAnnotation = 0,
    no_file_output: NoFileOutputAnnotation = False,
    steal_slurm: StealSlurmAnnotation = False,
    post_mortem: PostMortemAnnotation = False,
    output_to_console: OutputToConsoleAnnotation = False,
    worker_gpus: WorkerGPUsAnnotation = None,
    worker_cpus: WorkerCPUsAnnotation = None,
    worker_env: WorkerEnvAnnotation = None,
):
    """
    Fetch staged experiments from the database and run them (by default via Slurm).
    """
    start_experiments(
        ctx.obj['collection'],
        local=local,
        sacred_id=sacred_id,
        batch_id=batch_id,
        filter_dict=filter_dict,
        num_exps=num_exps,
        post_mortem=post_mortem,
        debug=debug,
        debug_server=debug_server,
        output_to_console=output_to_console,
        no_file_output=no_file_output,
        steal_slurm=steal_slurm,
        no_worker=no_worker,
        set_to_pending=True,
        worker_gpus=worker_gpus,
        worker_cpus=worker_cpus,
        worker_environment_vars=worker_env,
    )


@app.command('clean-jobs', rich_help_panel=_EXPERIMENTS, hidden=True)
@restrict_collection()
def clean_jobs_command(
    ctx: typer.Context,
    sacred_ids: Annotated[
        List[int],
        typer.Argument(
            help='Sacred IDs (_id in the database collection) of the experiments to claim.',
        ),
    ],
):
    """
    Cancel empty pending jobs.
    """
    cancel_empty_pending_jobs(ctx.obj['collection'], *sacred_ids)


@app.command('prepare-experiment', rich_help_panel=_EXPERIMENTS, hidden=True)
@restrict_collection()
def prepare_experiment_command(
    ctx: typer.Context,
    sacred_id: Annotated[
        int,
        typer.Option(
            '-id',
            '--sacred-id',
            help='Sacred ID (_id in the database collection) of the experiment. '
            'Takes precedence over other filters.',
        ),
    ],
    verbose: Annotated[
        bool,
        typer.Option(
            '-v',
            '--verbose',
            help='Whether to print debug messages.',
            is_flag=True,
        ),
    ] = False,
    unobserved: Annotated[
        bool,
        typer.Option(
            '-u',
            '--unobserved',
            help='Run the experiments without Sacred observers.',
            is_flag=True,
        ),
    ] = False,
    post_mortem: PostMortemAnnotation = False,
    stored_sources_dir: Annotated[
        Optional[str],
        typer.Option(
            '-ssd',
            '--stored-sources-dir',
            help='Load source files into this directory before starting.',
        ),
    ] = None,
    debug_server: DebugServerAnnotation = False,
):
    """
    Fetch experiment from database, prepare it and print the command to execute it.
    """
    prepare_experiment(
        ctx.obj['collection'],
        sacred_id,
        verbose,
        unobserved,
        post_mortem,
        stored_sources_dir,
        debug_server,
    )


@app.command('claim-experiment', rich_help_panel=_EXPERIMENTS, hidden=True)
@restrict_collection()
def claim_experiment_command(
    ctx: typer.Context,
    sacred_ids: Annotated[
        List[int],
        typer.Argument(
            help='Sacred IDs (_id in the database collection) of the experiments to claim.',
        ),
    ],
):
    """
    Claim an experiment from the database.
    """
    claim_experiment(ctx.obj['collection'], sacred_ids)


@app.command('launch-worker', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def launch_worker_command(
    ctx: typer.Context,
    num_exps: NumExperimentsAnnotation = 0,
    no_file_output: NoFileOutputAnnotation = False,
    steal_slurm: StealSlurmAnnotation = False,
    post_mortem: PostMortemAnnotation = False,
    debug: DebugAnnotation = False,
    debug_server: DebugServerAnnotation = False,
    output_to_console: OutputToConsoleAnnotation = False,
    worker_gpus: WorkerGPUsAnnotation = None,
    worker_cpus: WorkerCPUsAnnotation = None,
    worker_env: WorkerEnvAnnotation = None,
    sacred_id: SacredIdAnnotation = None,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
):
    """
    Launch a local worker that runs PENDING jobs.
    """
    start_experiments(
        ctx.obj['collection'],
        local=True,
        sacred_id=sacred_id,
        batch_id=batch_id,
        filter_dict=filter_dict,
        num_exps=num_exps,
        post_mortem=post_mortem,
        debug=debug,
        debug_server=debug_server,
        output_to_console=output_to_console,
        no_file_output=no_file_output,
        steal_slurm=steal_slurm,
        no_worker=False,
        set_to_pending=False,
        worker_gpus=worker_gpus,
        worker_cpus=worker_cpus,
        worker_environment_vars=worker_env,
    )


@app.command('print-fail-trace', rich_help_panel=_INFORMATION)
@restrict_collection()
def print_fail_trace_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [
        *States.FAILED,
        *States.KILLED,
        *States.INTERRUPTED,
    ],
    projection: ProjectionAnnotation = [],
):
    """
    Prints fail traces of all failed experiments.
    """
    print_fail_trace(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        batch_id=batch_id,
        filter_dict=filter_dict,
        projection=projection,
    )


@app.command('reload-sources', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def reload_sources_command(
    ctx: typer.Context,
    keep_old: Annotated[
        bool,
        typer.Option(
            '-k',
            '-keep-old',
            help='Keep the old source files in the database.',
            is_flag=True,
        ),
    ] = False,
    batch_ids: Annotated[
        Optional[List[int]],
        typer.Option(
            '-b',
            '--batch-ids',
            help='Batch IDs (batch_id in the database collection) of the experiments. '
            'Experiments that were staged together have the same batch_id.',
        ),
    ] = None,
    yes: YesAnnotation = False,
):
    """
    Reload stashed source files.
    """
    reload_sources(
        ctx.obj['collection'],
        batch_ids=batch_ids,
        keep_old=keep_old,
        yes=yes,
    )


@app.command('update-working-dir', rich_help_panel=_DATABASE)
@restrict_collection()
def update_working_dir_command(
    ctx: typer.Context,
    working_dir: Annotated[
        str,
        typer.Argument(
            help='The new working directory for the experiments.',
            exists=True,
            file_okay=False,
            dir_okay=True,
        ),
    ],
    batch_ids: Annotated[
        Optional[List[int]],
        typer.Option(
            '-b',
            '--batch-ids',
            help='Batch IDs (batch_id in the database collection) of the experiments. '
            'Experiments that were staged together have the same batch_id.',
        ),
    ] = None,
):
    """
    Change the working directory of experiments in case you moved the source code to a different location.
    """
    update_working_dir(
        ctx.obj['collection'],
        working_directory=working_dir,
        batch_ids=batch_ids,
    )


@app.command('print-command', rich_help_panel=_INFORMATION)
@restrict_collection()
def print_command_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = States.STAGED,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    num_exps: NumExperimentsAnnotation = 0,
    worker_gpus: WorkerGPUsAnnotation = None,
    worker_cpus: WorkerCPUsAnnotation = None,
    worker_env: WorkerEnvAnnotation = None,
    unresolved: Annotated[
        bool,
        typer.Option(
            '--unresolved',
            help='Whether to print the unresolved command.',
            is_flag=True,
        ),
    ] = False,
    no_interpolation: Annotated[
        bool,
        typer.Option(
            '--no-interpolation',
            help='Whether disable variable interpolation. Only compatible with --unresolved.',
            is_flag=True,
        ),
    ] = False,
):
    """
    Print the commands that would be executed by `start`.
    """
    print_command(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        batch_id=batch_id,
        filter_states=filter_states,
        filter_dict=filter_dict,
        num_exps=num_exps,
        worker_gpus=worker_gpus,
        worker_cpus=worker_cpus,
        worker_environment_vars=worker_env,
        unresolved=unresolved,
        resolve_interpolations=not no_interpolation,
    )


@app.command('print-experiment', rich_help_panel=_INFORMATION)
@restrict_collection()
def print_experiment_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = States.PENDING
    + States.STAGED
    + States.RUNNING
    + States.FAILED
    + States.KILLED
    + States.INTERRUPTED
    + States.COMPLETED,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    projection: ProjectionAnnotation = [],
):
    """
    Print the experiment document.
    """
    print_experiment(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        batch_id=batch_id,
        filter_dict=filter_dict,
        projection=projection,
    )


@app.command('print-output', rich_help_panel=_INFORMATION)
@restrict_collection()
def print_output_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = States.RUNNING
    + States.FAILED
    + States.KILLED
    + States.INTERRUPTED
    + States.COMPLETED,
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    slurm: Annotated[
        bool,
        typer.Option(
            '-sl',
            '--slurm',
            help='Whether to print the Slurm output instead of the experiment output.',
            is_flag=True,
        ),
    ] = False,
    head: Annotated[
        Optional[int],
        typer.Option(
            '-h',
            '--head',
            help='Print the first n lines of the output.',
        ),
    ] = None,
    tail: Annotated[
        Optional[int],
        typer.Option(
            '-t',
            '--tail',
            help='Print the last n lines of the output.',
        ),
    ] = None,
):
    """
    Print the output of experiments.
    """
    print_output(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        batch_id=batch_id,
        filter_dict=filter_dict,
        slurm=slurm,
        head=head,
        tail=tail,
    )


@app.command('reset', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def reset_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [
        *States.FAILED,
        *States.KILLED,
        *States.INTERRUPTED,
    ],
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    yes: YesAnnotation = False,
):
    """
    Reset the state of experiments by setting their state to STAGED and cleaning their database entry.
    Does not cancel Slurm jobs.
    """
    reset_experiments(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        batch_id=batch_id,
        filter_dict=filter_dict,
        yes=yes,
    )


@app.command('delete', rich_help_panel=_EXPERIMENTS)
@restrict_collection()
def delete_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [
        *States.STAGED,
        *States.FAILED,
        *States.KILLED,
        *States.INTERRUPTED,
    ],
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    no_cancel: Annotated[
        bool,
        typer.Option(
            '-nc',
            '--no-cancel',
            help='Do not cancel the experiments before deleting them.',
            is_flag=True,
        ),
    ] = False,
    yes: YesAnnotation = False,
):
    """
    Delete experiments by ID or state (cancels Slurm jobs first if not --no-cancel).
    """
    delete_experiments(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        batch_id=batch_id,
        filter_dict=filter_dict,
        yes=yes,
        cancel=not no_cancel,
    )
    get_db_collections.recompute_cache()


@app.command('drop', rich_help_panel=_DATABASE)
@restrict_collection(False)
def drop_command(
    ctx: typer.Context,
    pattern: Annotated[
        str, typer.Argument(help='A regex that must match the collections to print.')
    ] = r'.*',
    yes: YesAnnotation = False,
):
    """
    Drop collections from the database.

    Note: This is a dangerous operation and should only be used if you know what you are doing.
    """
    drop_collections(pattern=pattern, yes=yes)
    get_db_collections.recompute_cache()


@app.command('detect-killed', rich_help_panel=_DATABASE)
@restrict_collection()
def detect_killed_command(
    ctx: typer.Context,
):
    """
    Detect experiments where the corresponding Slurm jobs were killed externally.
    """
    detect_killed(ctx.obj['collection'])


@app.command('status', rich_help_panel=_INFORMATION)
@restrict_collection()
def status_command(
    ctx: typer.Context,
    update_status: UpdateStatusAnnotation = True,
    projection: ProjectionAnnotation = [],
):
    """
    Report status of experiments in the database collection.
    """
    print_status(
        ctx.obj['collection'], update_status=update_status, projection=projection
    )


@app.command('download-sources', rich_help_panel=_INFORMATION)
@restrict_collection()
def download_sources_command(
    ctx: typer.Context,
    target_directory: Annotated[
        str,
        typer.Argument(
            help='The directory where the source files should be restored.',
            exists=False,
            file_okay=False,
            dir_okay=True,
        ),
    ],
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [],
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
):
    """
    Download source files from the database to the provided path.
    """
    download_sources(
        target_directory,
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        filter_dict=filter_dict,
        batch_id=batch_id,
    )


@app.command('hold', rich_help_panel=_SLURM)
@restrict_collection()
def hold_command(
    ctx: typer.Context,
    batch_id: BatchIdAnnotation = None,
):
    """
    Hold queued experiments via SLURM.
    """
    hold_or_release_experiments(
        True,
        ctx.obj['collection'],
        batch_id=batch_id,
    )


@app.command('release', rich_help_panel=_SLURM)
@restrict_collection()
def release_command(
    ctx: typer.Context,
    batch_id: BatchIdAnnotation = None,
):
    """
    Release holded experiments via SLURM.
    """
    hold_or_release_experiments(
        False,
        ctx.obj['collection'],
        batch_id=batch_id,
    )


@app.command('queue', rich_help_panel=_INFORMATION)
@restrict_collection(False)
def queue_command(
    ctx: typer.Context,
    job_ids: List[str] = typer.Argument(
        help='The job IDs of the experiments to get the collection for.',
        default=None,
    ),
    filter_states: FilterStatesAnnotation = [*States.PENDING, *States.RUNNING],
    check_all: Annotated[
        bool,
        typer.Option(
            '-a',
            '--all',
            help='Whether to attempt finding the collection of the jobs of all users.',
            is_flag=True,
        ),
    ] = False,
    watch: Annotated[
        bool,
        typer.Option(
            '-w',
            '--watch',
            help='Whether to watch the queue.',
            is_flag=True,
        ),
    ] = False,
):
    """
    Prints the collections of the given job IDs. If none is specified, all jobs are considered.
    """
    print_queue(
        job_ids,
        filter_by_user=not check_all,
        filter_states=filter_states,
        watch=watch,
    )


app_description = typer.Typer(
    no_args_is_help=True,
    help='Manage descriptions of the experiments in a collection.',
    # chain=_AUTOCOMPLETE
)
app.add_typer(app_description, name='description', rich_help_panel=_EXPERIMENTS)


@app.command('detect-duplicates', rich_help_panel=_DATABASE)
@restrict_collection()
def detect_duplicates_command(
    ctx: typer.Context,
    filter_states: FilterStatesAnnotation = [
        *States.STAGED,
        *States.FAILED,
        *States.KILLED,
        *States.INTERRUPTED,
    ],
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
):
    """
    Prints duplicate experiment configurations.
    """
    print_duplicates(
        ctx.obj['collection'],
        filter_states=filter_states,
        filter_dict=filter_dict,
        batch_id=batch_id,
    )


@app_description.command('set')
@restrict_collection()
def description_set_command(
    ctx: typer.Context,
    description: Annotated[
        str,
        typer.Argument(
            help='The description to set.',
        ),
    ],
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [],
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    yes: YesAnnotation = False,
    no_resolve_description: NoResolveDescriptionAnnotation = False,
):
    """
    Sets the description of experiment(s).
    """
    collection_set_description(
        ctx.obj['collection'],
        description,
        sacred_id=sacred_id,
        filter_states=filter_states,
        filter_dict=filter_dict,
        batch_id=batch_id,
        yes=yes,
        resolve=not no_resolve_description,
    )


@app_description.command('delete')
@restrict_collection()
def description_delete_command(
    ctx: typer.Context,
    sacred_id: SacredIdAnnotation = None,
    filter_states: FilterStatesAnnotation = [],
    filter_dict: FilterDictAnnotation = None,
    batch_id: BatchIdAnnotation = None,
    yes: YesAnnotation = False,
):
    """
    Deletes the description of experiment(s).
    """
    collection_delete_description(
        ctx.obj['collection'],
        sacred_id=sacred_id,
        filter_states=filter_states,
        filter_dict=filter_dict,
        batch_id=batch_id,
        yes=yes,
    )


@app_description.command('list')
@restrict_collection()
def description_list_command(
    ctx: typer.Context, update_status: UpdateStatusAnnotation = False
):
    """
    Lists the descriptions of all experiments.
    """
    collection_list_descriptions(ctx.obj['collection'], update_status=update_status)


app_project = typer.Typer(
    no_args_is_help=True,
    help='Setting up new projects.',
)
app.add_typer(app_project, name='project')


@app_project.command('init')
@restrict_collection(False)
def init_project_command(
    ctx: typer.Context,
    directory: Annotated[
        str,
        typer.Argument(
            help='The directory in which to initialize the project.',
            exists=False,
            file_okay=False,
            dir_okay=True,
        ),
    ] = '.',
    template: Annotated[
        str,
        typer.Option(
            '-t',
            '--template',
            help='The template to use for the project. To view available templates use `seml project list-templates`.',
        ),
    ] = 'default',
    project_name: Annotated[
        Optional[str],
        typer.Option(
            '-n',
            '--name',
            help='The name of the project. (By default inferred from the directory name.)',
        ),
    ] = None,
    user_name: Annotated[
        Optional[str],
        typer.Option(
            '-u',
            '--username',
            help='The author name to use for the project. (By default inferred from $USER)',
        ),
    ] = None,
    user_mail: Annotated[
        Optional[str],
        typer.Option(
            '-m',
            '--usermail',
            help='The author email to use for the project. (By default empty.)',
        ),
    ] = None,
    git_remote: Annotated[
        Optional[str],
        typer.Option(
            '-r',
            '--git-remote',
            help='The git remote to use for the project. (By default SETTINGS.TEMPLATE_REMOTE.)',
        ),
    ] = None,
    git_commit: Annotated[
        Optional[str],
        typer.Option(
            '-c',
            '--git-commit',
            help='The exact git commit to use. May also be a tag or branch (By default latest)',
        ),
    ] = None,
    yes: YesAnnotation = False,
):
    """
    Initialize a new project in the given directory.
    """
    init_project(
        directory,
        project_name,
        user_name,
        user_mail,
        template,
        git_remote,
        git_commit,
        yes,
    )


@app_project.command('list-templates')
@restrict_collection(False)
def list_templates_command(
    ctx: typer.Context,
    git_remote: Annotated[
        Optional[str],
        typer.Option(
            '-r',
            '--git-remote',
            help='The git remote to use for the project. (By default SETTINGS.TEMPLATE_REMOTE.)',
        ),
    ] = None,
    git_commit: Annotated[
        Optional[str],
        typer.Option(
            '-c',
            '--git-commit',
            help='The exact git commit to use. May also be a tag or branch (By default latest)',
        ),
    ] = None,
):
    """
    List available project templates.
    """
    print_available_templates(git_remote, git_commit)


@dataclass
class CommandTreeNode:
    """Compact representation of the commands (and subtyper commands) of the app"""

    commands: Set[str]
    groups: Dict[str, 'CommandTreeNode']


@functools.lru_cache
def command_tree(app: typer.Typer) -> CommandTreeNode:
    return CommandTreeNode(
        commands={
            cmd.name if cmd.name else cmd.callback.__name__  # type: ignore
            for cmd in app.registered_commands
        },
        groups={
            (group.name if group.name else group.callback.__name__): command_tree(  # type: ignore
                group.typer_instance
            )
            for group in app.registered_groups
        },
    )


def split_args(
    args: List[str], command_tree: CommandTreeNode, combine: bool = True
) -> Tuple[List[List[str]], List[CommandTreeNode]]:
    split_cmd_args: List[List[str]] = [[]]
    cmd_stack = [command_tree]

    # Chaining is only allowed in the first level of the group hierarchy, so we only
    # split into a two level list
    for arg in args:
        if arg in cmd_stack[-1].groups:
            if len(cmd_stack) == 1:  # new subtyper at the top level
                split_cmd_args.append([arg])
                # chaining is allowed: stack[-1] may consume further commands after its child is done consuming
                cmd_stack.append(cmd_stack[-1].groups[arg])
            else:
                split_cmd_args[-1].append(arg)
                # no chaining below the first level: stack[-1] will not consume any more commands
                cmd_stack = cmd_stack[:-1] + [cmd_stack[-1].groups[arg]]
        elif arg in cmd_stack[-1].commands:
            if len(cmd_stack) == 1:  # new command at the top level
                split_cmd_args.append([arg])
            else:
                split_cmd_args[-1].append(arg)
                # no chaining below the first level: stack[-1] will not consume any more commands
                cmd_stack.pop()
        else:
            split_cmd_args[-1].append(arg)

    # Re-distribute shared args to each command in the first level of the hierarchy
    if len(split_cmd_args) == 1:
        return split_cmd_args, cmd_stack
    shared = split_cmd_args[0]
    chained_commands = split_cmd_args[1:]
    # If none of the shared args contains a collection
    # name, we add the default collection name.
    if all(arg.startswith('-') for arg in shared):
        shared.append('')
    # Combine commands with the shared arguments
    result = []
    for split in chained_commands:
        result.append(shared + split)

    return result, cmd_stack


def main():
    # We have to split the arguments manually to get proper chaining.
    # If we were to use typer built-in chaining, lists would end the chain.
    for args in split_args(sys.argv[1:], command_tree(app))[0]:
        # The app will typically exit after running once.
        # We want to run it multiple times, so we catch the SystemExit exception.
        from seml.console import console

        try:
            if len(args) >= 2:
                cmd = args[1]
            else:
                cmd = None
            with console.status(f'Running command: {cmd}'):
                app(args)
        except SystemExit as e:
            if e.code == 0:
                continue
            else:
                raise e


if __name__ == '__main__':
    main()


# If we are in autcompletion we must apply our parameter splitting
# to get correct autocompletion suggestions.
if AUTOCOMPLETING and os.environ.get('COMP_WORDS'):
    commands, stack = split_args(
        os.environ['COMP_WORDS'].split('\n'), command_tree(app)
    )
    cword = int(os.environ['COMP_CWORD'])
    if cword > 1:
        # Case where we complete a command
        # To find the right command, we must subtract the length of all previous commands
        # let's subtract -2 everywhere for seml <collection>
        cword -= 2
        for cmd in commands:
            cmd_length = len(cmd) - 2
            # We found our current command
            if cmd_length >= cword:
                break
            cword -= cmd_length
        cword += 2  # add back the -2 we subtracted above
    else:
        # If we complete collection names
        cmd = commands[0]
    os.environ['COMP_WORDS'] = '\n'.join(cmd)
    os.environ['COMP_CWORD'] = str(cword)
    # If we are not at the top level typer, we must not suggest top level commands
    # Note: `seml collection description list <tab><tab>` does not correctly autocomplete
    # as chaining is disabled on the app_description typer. However, if one were to enable
    # that its assumptions about chaining differs from our assumptions about chaining.
    app.info.chain = len(stack) == 1



import ast
import copy
import functools
import logging
import numbers
import os
import warnings
from itertools import combinations
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from seml.experiment.parameters import (
    cartesian_product_zipped_dict,
    generate_grid,
    sample_random_configs,
    zipped_dict,
)
from seml.experiment.sources import import_exe
from seml.settings import SETTINGS
from seml.utils import (
    Hashabledict,
    flatten,
    merge_dicts,
    remove_keys_from_nested,
    unflatten,
    working_directory,
)
from seml.utils.errors import ConfigError, ExecutableError

if TYPE_CHECKING:
    import sacred

RESERVED_KEYS = ['grid', 'fixed', 'random']


def unpack_config(config):
    config = convert_parameter_collections(config)
    children = {}
    reserved_dict = {}
    for key, value in config.items():
        if not isinstance(value, dict):
            continue

        if key not in RESERVED_KEYS:
            children[key] = value
        else:
            if key == 'random':
                if 'samples' not in value:
                    raise ConfigError(
                        'Random parameters must specify "samples", i.e. the number of random samples.'
                    )
                reserved_dict[key] = value
            else:
                reserved_dict[key] = value
    return reserved_dict, children


def extract_parameter_set(input_config: dict, key: str):
    flattened_dict = flatten(input_config.get(key, {}))
    keys = flattened_dict.keys()
    if key != 'fixed':
        keys = [
            '.'.join(k.split('.')[:-1])
            for k in keys
            if flattened_dict[k] != 'parameter_collection'
        ]
    return set(keys)


def convert_parameter_collections(input_config: dict):
    flattened_dict = flatten(input_config)
    parameter_collection_keys = [
        k for k in flattened_dict.keys() if flattened_dict[k] == 'parameter_collection'
    ]
    if len(parameter_collection_keys) > 0:
        logging.warning(
            'Parameter collections are deprecated. Use dot-notation for nested parameters instead.'
        )
    while len(parameter_collection_keys) > 0:
        k = parameter_collection_keys[0]
        del flattened_dict[k]
        # sub1.sub2.type ==> # sub1.sub2
        k = '.'.join(k.split('.')[:-1])
        parameter_collections_params = [
            param_key for param_key in flattened_dict.keys() if param_key.startswith(k)
        ]
        for p in parameter_collections_params:
            if f'{k}.params' in p:
                new_key = p.replace(f'{k}.params', k)
                if new_key in flattened_dict:
                    raise ConfigError(
                        f'Could not convert parameter collections due to key collision: {new_key}.'
                    )
                flattened_dict[new_key] = flattened_dict[p]
                del flattened_dict[p]
        parameter_collection_keys = [
            k
            for k in flattened_dict.keys()
            if flattened_dict[k] == 'parameter_collection'
        ]
    return unflatten(flattened_dict)


def standardize_config(config: dict):
    config = unflatten(flatten(config), levels=[0])
    out_dict = {}
    for k in RESERVED_KEYS:
        if k == 'fixed':
            out_dict[k] = config.get(k, {})
        else:
            out_dict[k] = unflatten(config.get(k, {}), levels=[-1])
    return out_dict


def invert_config(config: dict):
    reserved_sets = [(k, set(config.get(k, {}).keys())) for k in RESERVED_KEYS]
    inverted_config = {}
    for k, params in reserved_sets:
        for p in params:
            L = inverted_config.get(p, [])
            L.append(k)
            inverted_config[p] = L
    return inverted_config


def detect_duplicate_parameters(
    inverted_config: dict, sub_config_name: str = None, ignore_keys: dict = None
):
    if ignore_keys is None:
        ignore_keys = {'random': ('seed', 'samples')}

    duplicate_keys = []
    for p, L in inverted_config.items():
        if len(L) > 1:
            if 'random' in L and p in ignore_keys['random']:
                continue
            duplicate_keys.append((p, L))

    if len(duplicate_keys) > 0:
        if sub_config_name:
            raise ConfigError(
                f'Found duplicate keys in sub-config {sub_config_name}: '
                f'{duplicate_keys}'
            )
        else:
            raise ConfigError(f'Found duplicate keys: {duplicate_keys}')

    start_characters = {x[0] for x in inverted_config.keys()}
    buckets = {
        k: {x for x in inverted_config.keys() if x.startswith(k)}
        for k in start_characters
    }

    if sub_config_name:
        error_str = (
            f'Conflicting parameters in sub-config {sub_config_name}, most likely '
            'due to ambiguous use of dot-notation in the config dict. Found '
            "parameter '{p1}' in dot-notation starting with other parameter "
            "'{p2}', which is ambiguous."
        )
    else:
        error_str = (
            'Conflicting parameters, most likely '
            'due to ambiguous use of dot-notation in the config dict. Found '
            "parameter '{p1}' in dot-notation starting with other parameter "
            "'{p2}', which is ambiguous."
        )

    for k in buckets.keys():
        for p1, p2 in combinations(buckets[k], r=2):
            if p1.startswith(
                f'{p2}.'
            ):  # with "." after p2 to catch cases like "test" and "test1", which are valid.
                raise ConfigError(error_str.format(p1=p1, p2=p2))
            elif p2.startswith(f'{p1}.'):
                raise ConfigError(error_str.format(p1=p1, p2=p2))


def generate_configs(experiment_config, overwrite_params=None):
    """Generate parameter configurations based on an input configuration.

    Input is a nested configuration where on each level there can be 'fixed', 'grid', and 'random' parameters.

    In essence, we take the cartesian product of all the `grid` parameters and take random samples for the random
    parameters. The nested structure makes it possible to define different parameter spaces e.g. for different datasets.
    Parameter definitions lower in the hierarchy overwrite parameters defined closer to the root.

    For each leaf configuration we take the maximum of all num_samples values on the path since we need to have the same
    number of samples for each random parameter.

    For each configuration of the `grid` parameters we then create `num_samples` configurations of the random
    parameters, i.e. leading to `num_samples * len(grid_configurations)` configurations.

    See Also `examples/example_config.yaml` and the example below.

    Parameters
    ----------
    experiment_config: dict
        Dictionary that specifies the "search space" of parameters that will be enumerated. Should be
        parsed from a YAML file.
    overwrite_params: Optional[dict]
        Flat dictionary that overwrites configs. Resulting duplicates will be removed.

    Returns
    -------
    all_configs: list of dicts
        Contains the individual combinations of the parameters.


    """

    reserved, next_level = unpack_config(experiment_config)
    reserved = standardize_config(reserved)
    if not any([len(reserved.get(k, {})) > 0 for k in RESERVED_KEYS]):
        raise ConfigError(
            'No parameters defined under grid, fixed, or random in the config file.'
        )
    level_stack = [('', next_level)]
    config_levels = [reserved]
    final_configs = []

    detect_duplicate_parameters(invert_config(reserved), None)

    while len(level_stack) > 0:
        current_sub_name, sub_vals = level_stack.pop(0)
        sub_config, sub_levels = unpack_config(sub_vals)
        if current_sub_name != '' and not any(
            [len(sub_config.get(k, {})) > 0 for k in RESERVED_KEYS]
        ):
            raise ConfigError(
                f'No parameters defined under grid, fixed, or random in sub-config {current_sub_name}.'
            )
        sub_config = standardize_config(sub_config)
        config_above = config_levels.pop(0)

        inverted_sub_config = invert_config(sub_config)
        detect_duplicate_parameters(inverted_sub_config, current_sub_name)

        inverted_config_above = invert_config(config_above)
        redefined_parameters = set(inverted_sub_config.keys()).intersection(
            set(inverted_config_above.keys())
        )

        if len(redefined_parameters) > 0:
            logging.info(
                f"Found redefined parameters in sub-config '{current_sub_name}': {redefined_parameters}. "
                f'Definitions in sub-configs override more general ones.'
            )
            config_above = copy.deepcopy(config_above)
            for p in redefined_parameters:
                sections = inverted_config_above[p]
                for s in sections:
                    del config_above[s][p]

        config = merge_dicts(config_above, sub_config)

        if len(sub_levels) == 0:
            final_configs.append((current_sub_name, config))

        for sub_name, sub_vals in sub_levels.items():
            new_sub_name = (
                f'{current_sub_name}.{sub_name}' if current_sub_name != '' else sub_name
            )
            level_stack.append((new_sub_name, sub_vals))
            config_levels.append(config)

    all_configs = []
    for subconfig_name, conf in final_configs:
        conf = standardize_config(conf)
        random_params = conf.get('random', {})
        fixed_params = flatten(conf.get('fixed', {}))
        grid_params = conf.get('grid', {})

        if len(random_params) > 0:
            num_samples = random_params['samples']
            root_seed = random_params.get('seed', None)
            random_sampled = sample_random_configs(
                flatten(random_params), seed=root_seed, samples=num_samples
            )

        grids = [generate_grid(v, parent_key=k) for k, v in grid_params.items()]
        grid_configs = dict([sub for item in grids for sub in item])
        grouped_configs = zipped_dict(grid_configs)
        grid_product = list(cartesian_product_zipped_dict(grouped_configs))

        with_fixed = [{**d, **fixed_params} for d in grid_product]
        if len(random_params) > 0:
            with_random = [
                {**grid, **random} for grid in with_fixed for random in random_sampled
            ]
        else:
            with_random = with_fixed
        all_configs.extend(with_random)

    # Cast NumPy integers to normal integers since PyMongo doesn't like them
    all_configs = [
        {
            k: int(v)
            if isinstance(v, numbers.Integral) and not isinstance(v, bool)
            else v
            for k, v in config.items()
        }
        for config in all_configs
    ]

    if overwrite_params is not None:
        all_configs = [merge_dicts(config, overwrite_params) for config in all_configs]
        base_length = len(all_configs)
        # We use a dictionary instead a set because dictionary keys are ordered as of Python 3
        all_configs = list({Hashabledict(**config): None for config in all_configs})
        new_length = len(all_configs)
        if base_length != new_length:
            diff = base_length - new_length
            logging.warning(
                f'Parameter overwrite caused {diff} identical configs. Duplicates were removed.'
            )

    all_configs = [unflatten(conf) for conf in all_configs]
    return all_configs


def generate_named_config(named_config_dict: Dict) -> List[str]:
    """Generates a sequence of named configs that is resolved by sacred in-order

    Parameters
    ----------
    named_config_dict : Dict
        Flattened configuration before parsing the named configurations.

    Returns
    -------
    List[str]
        A sequence of named configuration in the order that is defined by the `named_config_dict` input
    """
    # Parse named config names and priorities
    names, priorities = {}, {}
    for k, v in named_config_dict.items():
        if k.startswith(SETTINGS.NAMED_CONFIG.PREFIX):
            if isinstance(v, str):
                v = dict(name=str(v))
            if not isinstance(v, Dict):
                raise ConfigError(
                    f'Named configs must be given as '
                    f'{SETTINGS.NAMED_CONFIG.PREFIX}{"{identifier}"}: str | '
                    '{"name": str, "priority": int}'
                )
            for attribute, value in v.items():
                if attribute == SETTINGS.NAMED_CONFIG.KEY_NAME:
                    if not isinstance(value, str):
                        raise ConfigError(
                            f'Named config names should be strings, not {value} ({value.__class__})'
                        )
                    names[k] = value
                elif attribute == SETTINGS.NAMED_CONFIG.KEY_PRIORITY:
                    try:
                        value = int(value)
                    except (ValueError, TypeError):
                        raise ConfigError(
                            f'Named config priorities should be non-negative integers, not {value} ({value.__class__})'
                        )
                    priorities[k] = value
                else:
                    raise ConfigError(
                        f'Named configs only have the attributes {[SETTINGS.NAMED_CONFIG.KEY_NAME, SETTINGS.NAMED_CONFIG.KEY_PRIORITY]}'
                    )
    for idx in priorities:
        if idx not in names:
            raise ConfigError(
                f'Defined a priority but not a name for named config {idx}'
            )
    return [
        names[idx]
        for idx in sorted(
            names, key=lambda idx: (priorities.get(idx, float('inf')), names[idx])
        )
    ]


def generate_named_configs(configs: List[Dict]) -> Tuple[List[Dict], List[List[str]]]:
    """From experiment configurations, generates both the config updates as well as the named configs in the order specified.

    Parameters
    ----------
    configs : List[Dict]
        Input configurations.

    Returns
    -------
    List[Dict]
        For each input configuration, the output configuration that does not contain named configuration specifiers anymore.

    List[List[str]]]
        For each input configuration, the sequence of named configurations in the order specified.
    """
    result_configs, result_named_configs = [], []
    for config in configs:
        result_configs.append(
            {
                k: v
                for k, v in config.items()
                if not k.startswith(SETTINGS.NAMED_CONFIG.PREFIX)
            }
        )
        result_named_configs.append(generate_named_config(config))
    return result_configs, result_named_configs


@functools.lru_cache
def load_config_dict(cfg_name: str):
    """
    Wrapper around sacred internal function to load a configuration file.
    This wrapper is cached to avoid loading the same file multiple times.

    Parameters
    ----------
    cfg_name : str
        Path to the configuration file.

    Returns
    -------
    ConfigDict
        The configuration dictionary.
    """
    from sacred.config.config_dict import ConfigDict
    from sacred.config.config_files import load_config_file

    return ConfigDict(load_config_file(cfg_name))


_SCAFFOLD_KEYS = (
    'config_updates',
    'named_configs_to_use',
    'config',
    'fallback',
    'presets',
    'fixture',
    'logger',
    'seed',
    'rnd',
    'config_mods',
    'summaries',
)


def _get_scaffold_state(scaffolding):
    return {
        k2: {k: getattr(scaffold, k) for k in _SCAFFOLD_KEYS}
        for k2, scaffold in scaffolding.items()
    }


def _set_scaffold_state(scaffolding, state):
    from copy import copy

    for k2, scaffold in scaffolding.items():
        for k, v in state[k2].items():
            setattr(scaffold, k, copy(v))


def _sacred_create_configs(
    exp: 'sacred.Experiment',
    configs: List[Dict],
    named_configs: Optional[List[Tuple[str]]] = None,
) -> List[Dict]:
    """Creates configs from an experiment and update values. This is done by re-implementing sacreds `sacred.initialize.create_run`
    method. Doing this is significantly faster, but it can be out-of-sync with sacred's current implementation.

    Parameters
    ----------
    exp : sacred.Experiment
        The sacred experiment to create configs for
    configs : List[Dict]
        Configuration updates for each experiment
    named_configs : Optional[List[Tuple[str]]], optional
        Named configs for each experiment, by default ()

    Returns
    -------
    Dict
        The updated configurations containing all values derived from the experiment.
    """
    from copy import deepcopy

    from sacred.config.utils import undogmatize
    from sacred.initialize import (
        Scaffold,
        create_scaffolding,
        distribute_config_updates,
        distribute_presets,
        gather_ingredients_topological,
        get_configuration,
        get_scaffolding_and_config_name,
    )
    from sacred.utils import (
        convert_to_nested_dict,
        iterate_flattened,
        join_paths,
        recursive_update,
        set_by_dotted_path,
    )

    from seml.console import track

    def run_named_config(scaffold: Scaffold, cfg_name: str):
        # This version of sacred.initialize.Scaffold.run_named_config uses our
        # cached version of load_config_dict. This is necessary to avoid loading the same file multiple times.
        if os.path.isfile(cfg_name):
            nc = load_config_dict(cfg_name)
            cfg = nc(
                fixed=scaffold.get_config_updates_recursive(),
                preset=scaffold.presets,
                fallback=scaffold.fallback,
            )
            return undogmatize(cfg)
        return scaffold.run_named_config(cfg_name)

    configs_resolved = []
    if named_configs is None:
        named_configs = [()] * len(configs)
    sorted_ingredients = gather_ingredients_topological(exp)
    scaffolding = create_scaffolding(exp, sorted_ingredients)
    init_state = deepcopy(_get_scaffold_state(scaffolding))
    # get all split non-empty prefixes sorted from deepest to shallowest
    prefixes = sorted(
        [s.split('.') for s in scaffolding if s != ''],
        reverse=True,
        key=lambda p: len(p),
    )

    for config, named_config in track(
        list(zip(configs, named_configs)),
        description='Resolving configurations',
        disable=len(configs) < SETTINGS.CONFIG_RESOLUTION_PROGRESS_BAR_THRESHOLD,
    ):
        # The following code is adapted from sacred directly: This results in a significant speedup
        # as we only care about the config but not about creating runs, however it is more error prone
        # to changes to sacred
        _set_scaffold_state(scaffolding, init_state)
        # --------- configuration process -------------------
        # Phase 1: Config updates
        config_updates = convert_to_nested_dict(config)
        distribute_config_updates(prefixes, scaffolding, config_updates)

        # Phase 2: Named Configs
        for ncfg in named_config:
            scaff, cfg_name = get_scaffolding_and_config_name(ncfg, scaffolding)
            scaff.gather_fallbacks()
            ncfg_updates = run_named_config(scaff, cfg_name)
            distribute_presets(scaff.path, prefixes, scaffolding, ncfg_updates)
            for ncfg_key, value in iterate_flattened(ncfg_updates):
                set_by_dotted_path(
                    config_updates, join_paths(scaff.path, ncfg_key), value
                )

        distribute_config_updates(prefixes, scaffolding, config_updates)

        # Phase 3: Normal config scopes
        for scaffold in scaffolding.values():
            scaffold.gather_fallbacks()
            scaffold.set_up_config()

            # update global config
            config = get_configuration(scaffolding)
            # run config hooks
            config_hook_updates = scaffold.run_config_hooks(
                config, exp.default_command, None
            )
            recursive_update(scaffold.config, config_hook_updates)

        # Phase 4: finalize seeding
        for scaffold in reversed(list(scaffolding.values())):
            scaffold.set_up_seed()  # partially recursive

        config_resolved = get_configuration(scaffolding)
        configs_resolved.append(
            remove_keys_from_nested(
                config_resolved, config_get_exclude_keys(config_resolved, config)
            )
        )
    return configs_resolved


def resolve_configs(
    executable: str,
    conda_env: str,
    configs: List[Dict],
    named_configs: List[List[str]],
    working_dir: str,
) -> List[Dict]:
    """Resolves configurations by adding keys that are only added when the experiment is run to the MongoDB

    Parameters
    ----------
    executable : str
        Path to the executable
    conda_env : str
        Which conda environment to use
    configs : List[Dict]
        All experiment configurations
    named_configs : List[str]
        For each experiment, the named configurations to use.
    working_dir : str
        Which working directory to use

    Returns
    -------
    List[Dict]
        Resolved configurations
    """
    import sacred

    from seml.experiment.experiment import Experiment

    exp_module = import_exe(executable, conda_env, working_dir)

    # Extract experiment from module
    exps = [
        v
        for k, v in exp_module.__dict__.items()
        if isinstance(v, (sacred.Experiment, Experiment))
    ]
    if len(exps) == 0:
        raise ExecutableError(
            f"Found no Sacred experiment. Something is wrong in '{executable}'."
        )
    elif len(exps) > 1:
        raise ExecutableError(
            f"Found more than 1 Sacred experiment in '{executable}'. "
            f"Can't resolve configs."
        )
    exp = exps[0]
    if not isinstance(exp, Experiment):
        logging.warning(
            'The use of sacred.Experiment is deprecated. Please use seml.Experiment instead.\n'
            'seml.Experiment already includes typical MongoDB observer and logging setups.\n'
            'Please familiar yourself with the new API and adjust your code accordingly.\n'
            'See https://github.com/TUM-DAML/seml/blob/master/examples/example_experiment.py'
        )
    with working_directory(working_dir):
        return _sacred_create_configs(exp, configs, named_configs)


def check_config(
    executable: str, conda_env: str, configs: List[Dict], working_dir: str
):
    """Check if the given configs are consistent with the Sacred experiment in the given executable.

    Parameters
    ----------
    executable: str
        The Python file containing the experiment.
    conda_env: str
        The experiment's Anaconda environment.
    configs: List[Dict]
        Contains the parameter configurations.
    working_dir : str
        The current working directory.
    """
    import sacred

    exp_module = import_exe(executable, conda_env, working_dir)

    # Extract experiment from module
    exps = [
        v for k, v in exp_module.__dict__.items() if isinstance(v, sacred.Experiment)
    ]
    if len(exps) == 0:
        raise ExecutableError(
            f"Found no Sacred experiment. Something is wrong in '{executable}'."
        )
    elif len(exps) > 1:
        raise ExecutableError(
            f"Found more than 1 Sacred experiment in '{executable}'. "
            f"Can't check parameter configs. Disable via --no-sanity-check."
        )
    exp = exps[0]

    empty_run = sacred.initialize.create_run(
        exp, exp.default_command, config_updates=None, named_configs=()
    )

    captured_args = {
        sacred.utils.join_paths(cf.prefix, n)
        for cf in exp.captured_functions
        for n in cf.signature.arguments
    }

    config_keys_empty_run = set(flatten(empty_run.config).keys())

    for config in configs:
        config_flat = flatten(config)
        config_keys_added = set(config_flat.keys()).difference(config_keys_empty_run)

        # Check for unused arguments
        for conf in sorted(config_keys_added):
            if not (set(sacred.utils.iter_prefixes(conf)) & captured_args):
                raise sacred.utils.ConfigAddedError(
                    conf,
                    config={
                        k: v for k, v in config_flat.items() if k in config_keys_added
                    },
                )

        # Check for missing arguments
        options = empty_run.config.copy()
        options.update(config)
        options.update({k: None for k in sacred.utils.ConfigAddedError.SPECIAL_ARGS})
        empty_run.main_function.signature.construct_arguments((), {}, options, False)


def restore(flat):
    """
    Restore more complex data that Python's json can't handle (e.g. Numpy arrays).
    Copied from sacred.serializer for performance reasons.
    """
    import json

    import jsonpickle

    return jsonpickle.decode(json.dumps(flat), keys=True)


def _convert_value(value):
    """
    Parse string as python literal if possible and fallback to string.
    Copied from sacred.arg_parser for performance reasons.
    """

    try:
        return restore(ast.literal_eval(value))
    except (ValueError, SyntaxError):
        # use as string if nothing else worked
        return value


def convert_values(val):
    if isinstance(val, dict):
        for key, inner_val in val.items():
            val[key] = convert_values(inner_val)
    elif isinstance(val, list):
        for i, inner_val in enumerate(val):
            val[i] = convert_values(inner_val)
    elif isinstance(val, str):
        return _convert_value(val)
    return val


def read_config(config_path: Union[str, Path]):
    import yaml

    from seml import __version__
    from seml.utils.yaml import YamlUniqueLoader

    with open(config_path) as conf:
        config_dict = convert_values(yaml.load(conf, Loader=YamlUniqueLoader))

    if 'seml' not in config_dict:
        raise ConfigError("Please specify a 'seml' dictionary.")

    seml_dict = config_dict['seml']
    del config_dict['seml']

    for k in seml_dict.keys():
        if k not in SETTINGS.VALID_SEML_CONFIG_VALUES:
            raise ConfigError(f'{k} is not a valid value in the `seml` config block.')

    if SETTINGS.SEML_CONFIG_VALUE_VERSION in seml_dict:
        raise ConfigError(
            f'Using {SETTINGS.SEML_CONFIG_VALUE_VERSION} in the `seml` config block is prohibited.'
        )

    version_array = [(int(x) if x.isdecimal() else x) for x in __version__.split('.')]
    seml_dict[SETTINGS.SEML_CONFIG_VALUE_VERSION] = version_array

    determine_executable_and_working_dir(config_path, seml_dict)

    # Get list of slurm configs
    slurm_list = config_dict.get('slurm', [])
    del config_dict['slurm']

    if slurm_list is None:
        slurm_list = []

    # Check for deprecated `slurm` dictionary
    if isinstance(slurm_list, dict):
        warnings.warn('`slurm` is expected to be a list of slurm configurations.')
        slurm_list = [slurm_list]

    # Sanity check
    for slurm_conf in slurm_list:
        for k in slurm_conf.keys():
            if k not in SETTINGS.VALID_SLURM_CONFIG_VALUES:
                raise ConfigError(
                    f'{k} is not a valid value in the `slurm` config block.'
                )
            if k == 'sbatch_options' and slurm_conf['sbatch_options'] is None:
                slurm_conf['sbatch_options'] = {}

    # If we have no config, we should add one
    if len(slurm_list) == 0:
        slurm_list.append({})

    return seml_dict, slurm_list, config_dict


def determine_executable_and_working_dir(config_path, seml_dict):
    """
    Determine the working directory of the project and chdir into the working directory.
    Parameters
    ----------
    config_path: Path to the config file
    seml_dict: SEML config dictionary

    Returns
    -------
    None
    """
    config_dir = str(Path(config_path).expanduser().resolve().parent)
    working_dir = config_dir
    if 'executable' not in seml_dict:
        raise ConfigError('Please specify an executable path for the experiment.')
    executable = seml_dict['executable']
    with working_directory(working_dir):
        executable_relative_to_config = os.path.exists(executable)
    executable_relative_to_project_root = False
    if 'project_root_dir' in seml_dict:
        with working_directory(config_dir):
            working_dir = str(
                Path(seml_dict['project_root_dir']).expanduser().resolve()
            )
        seml_dict['use_uploaded_sources'] = True
        with working_directory(working_dir):  # use project root as base dir from now on
            executable_relative_to_project_root = os.path.exists(executable)
        del seml_dict['project_root_dir']  # from now on we use only the working dir
    else:
        seml_dict['use_uploaded_sources'] = False
        logging.warning(
            "'project_root_dir' not defined in seml config. Source files will not be saved in MongoDB."
        )
    seml_dict['working_dir'] = working_dir
    if not (executable_relative_to_config or executable_relative_to_project_root):
        raise ExecutableError('Could not find the executable.')
    with working_directory(working_dir):
        executable = str(Path(executable).expanduser().resolve())
        if executable_relative_to_project_root:
            seml_dict['executable'] = str(Path(executable).relative_to(working_dir))
        else:
            seml_dict['executable'] = str(Path(executable).relative_to(config_dir))

        if 'output_dir' in seml_dict:
            seml_dict['output_dir'] = str(
                Path(seml_dict['output_dir']).expanduser().resolve()
            )


def remove_prepended_dashes(param_dict):
    new_dict = {}
    for k, v in param_dict.items():
        if k.startswith('--'):
            new_dict[k[2:]] = v
        elif k.startswith('-'):
            new_dict[k[1:]] = v
        else:
            new_dict[k] = v
    return new_dict


def config_get_exclude_keys(config: Dict, config_unresolved: Dict) -> List[str]:
    """Gets the key that should be excluded from identifying a config. These should
    e.g. not be used in hashing

    Parameters
    ----------
    config : Dict
        the configuration after resolution by sacred
    config_unresolved : Dict
        the configuration before resolution by sacred

    Returns
    -------
    List[str]
        keys that do not identify the config
    """
    exclude_keys = SETTINGS.CONFIG_EXCLUDE_KEYS
    if SETTINGS.CONFIG_KEY_SEED not in config_unresolved:
        # The seed will only be included (e.g. for hashing) if explicited in the unresolved configuration
        exclude_keys.append(SETTINGS.CONFIG_KEY_SEED)
    return exclude_keys


def requires_interpolation(
    document: Dict,
    allow_interpolation_keys: List[str] = SETTINGS.ALLOW_INTERPOLATION_IN,
) -> bool:
    r"""
    Check if a document requires variable interpolation. This is done by checking if
    any value matches the regex: .*\${.+}.*

    Parameters
    ----------
    document : Dict
        The document to check
    allow_interpolation_keys : List[str]
        All keys that should be permitted to do variable interpolation. Other keys are taken from the unresolved config.

    Returns
    -------
    bool
        True if the document requires variable interpolation
    """
    import re

    flat_dict = flatten(document)
    pattern = re.compile(r'.*\${.+}.*')

    def check_interpolation(key, value):
        if not any(
            key.startswith(allowed_key) for allowed_key in allow_interpolation_keys
        ):
            return False
        return isinstance(value, str) and pattern.match(value) is not None

    return any(map(check_interpolation, flat_dict.keys(), flat_dict.values()))


def resolve_interpolations(
    document: Dict,
    allow_interpolation_keys: List[str] = SETTINGS.ALLOW_INTERPOLATION_IN,
) -> Dict:
    """Resolves variable interpolation using `OmegaConf`

    Parameters
    ----------
    documents : Dict
        The document to resolve.
    allow_interpolations_in : List[str]
        All keys that should be permitted to do variable interpolation. Other keys are taken from the unresolved config.

    Returns
    -------
    Dict
        The resolved document
    """
    if not requires_interpolation(document, allow_interpolation_keys):
        return document

    from omegaconf import OmegaConf

    resolved = OmegaConf.to_container(
        OmegaConf.create(document, flags={'allow_objects': True}), resolve=True
    )
    resolved_flat = {
        key: value
        for key, value in flatten(resolved, sep='.').items()
        if any(key.startswith(allowed_key) for allowed_key in allow_interpolation_keys)
    }
    unresolved_flat = {
        key: value
        for key, value in flatten(document, sep='.').items()
        if not any(
            key.startswith(allowed_key) for allowed_key in allow_interpolation_keys
        )
    }
    assert resolved_flat.keys().isdisjoint(
        unresolved_flat.keys()
    ), f'Overlap between unresolved and resolved dicts: {resolved_flat.keys().intersection(unresolved_flat.keys())}'
    resolved = unflatten({**resolved_flat, **unresolved_flat})
    return resolved



import logging
import subprocess
from typing import Optional

from seml.commands.manage import detect_killed
from seml.database import build_filter_dict, get_collection
from seml.settings import SETTINGS
from seml.utils import s_if


def hold_or_release_experiments(
    hold: bool,
    db_collection_name: str,
    batch_id: Optional[int] = None,
):
    """
    Holds or releases experiments that are currently in the SLURM queue.

    Parameters
    ----------
    hold : bool
        Whether to hold or release the experiments
    db_collection_name : str
        The collection to hold or release experiments from
    batch_id : Optional[int], optional
        Filter on the batch ID of experiments, by default None
    """
    import shlex

    detect_killed(db_collection_name, False)

    filter_dict = build_filter_dict([*SETTINGS.STATES.PENDING], batch_id, None, None)
    collection = get_collection(db_collection_name)
    experiments = list(collection.find(filter_dict, {'slurm': 1}))

    arrays = set()
    for exp in experiments:
        for slurm in exp['slurm']:
            if (
                'array_id' not in slurm
            ):  # Skip experiments that are not in the SLURM queue
                continue
            arrays.add(slurm['array_id'])

    opteration = 'hold' if hold else 'release'
    subprocess.run(
        f'scontrol {opteration} {shlex.quote(" ".join(map(str, arrays)))}',
        shell=True,
        check=True,
        stdout=subprocess.DEVNULL,
    )
    # User feedback
    op_name = 'Held' if hold else 'Released'
    n_exp = len(experiments)
    n_job = len(arrays)
    logging.info(
        f'{op_name} {n_exp} experiment{s_if(n_exp)} in {n_job} job{s_if(n_job)}.'
    )



import copy
import logging
import math
import os
import shutil
import subprocess
import sys
import time
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence

from seml.commands.manage import cancel_experiment_by_id, reset_slurm_dict
from seml.database import build_filter_dict, get_collection
from seml.experiment.command import (
    get_command_from_exp,
    get_environment_variables,
    get_shell_command,
)
from seml.experiment.sources import load_sources_from_db
from seml.settings import SETTINGS
from seml.utils import (
    assert_package_installed,
    find_jupyter_host,
    load_text_resource,
    s_if,
)
from seml.utils.errors import ArgumentError, ConfigError
from seml.utils.slurm import (
    get_cluster_name,
    get_current_slurm_array_id,
    get_current_slurm_job_id,
    get_slurm_jobs,
)

if TYPE_CHECKING:
    from pymongo.collection import Collection

States = SETTINGS.STATES
SlurmStates = SETTINGS.SLURM_STATES


def get_output_dir_path(config):
    output_dir = config['seml'].get('output_dir', '.')
    output_dir_path = str(Path(output_dir).expanduser().resolve())
    if not os.path.isdir(output_dir_path):
        raise ConfigError(f"Output directory '{output_dir_path}' does not exist.")
    return output_dir_path


def get_exp_name(exp_config, db_collection_name):
    if 'name' in exp_config['seml']:
        name = exp_config['seml']['name']
    else:
        name = db_collection_name
    return name


def set_slurm_job_name(
    sbatch_options: Dict[str, Any],
    name: str,
    exp: Dict,
    db_collection_name: str,
):
    if 'job-name' in sbatch_options:
        raise ConfigError(
            "Can't set sbatch `job-name` parameter explicitly. "
            'Use `name` parameter instead and SEML will do that for you.'
        )
    job_name = f"{name}_{exp['batch_id']}"
    sbatch_options['job-name'] = job_name
    if sbatch_options.get('comment', db_collection_name) != db_collection_name:
        raise ConfigError(
            "Can't set sbatch `comment` parameter explicitly. "
            'SEML will do that for you and set it to the collection name.'
        )
    sbatch_options['comment'] = db_collection_name


def create_slurm_options_string(slurm_options: dict, srun: bool = False):
    """
    Convert a dictionary with sbatch_options into a string that can be used in a bash script.

    Parameters
    ----------
    slurm_options: Dictionary containing the sbatch options.
    srun: Construct options for an srun command instead of an sbatch script.

    Returns
    -------
    slurm_options_str: sbatch option string.
    """
    if srun:
        option_structure = ' {prepend}{key}={value}'
    else:
        option_structure = '#SBATCH {prepend}{key}={value}\n'

    slurm_options_str = ''
    for key, value_raw in slurm_options.items():
        prepend = '-' if len(key) == 1 else '--'
        if key in ['partition', 'p'] and isinstance(value_raw, list):
            value = ','.join(value_raw)
        else:
            value = value_raw
        slurm_options_str += option_structure.format(
            prepend=prepend, key=key, value=value
        )
    return slurm_options_str


def start_sbatch_job(
    collection: 'Collection',
    exp_array: Sequence[Dict],
    slurm_options_id: int,
    sbatch_options: Dict,
    unobserved: bool = False,
    name: Optional[str] = None,
    output_dir_path: str = '.',
    max_simultaneous_jobs: Optional[int] = None,
    experiments_per_job: int = 1,
    debug_server: bool = False,
):
    """Run a list of experiments as a job on the Slurm cluster.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    exp_array: List[dict]
        List of chunks of experiments to run. Each chunk is a list of experiments.
    unobserved: bool
        Disable all Sacred observers (nothing written to MongoDB).
    name: str
        Job name, used by Slurm job and output file.
    output_dir_path: str
        Directory (relative to home directory) where to store the slurm output files.
    sbatch_options: dict
        A dictionary that contains options for #SBATCH, e.g. {'mem': 8000} to limit the job's memory to 8,000 MB.
    max_simultaneous_jobs: int
        Maximum number of Slurm jobs running simultaneously.
    debug_server: bool
        Run jobs with a debug server.

    Returns
    -------
    None
    """
    from tempfile import NamedTemporaryFile

    seml_conf = exp_array[0]['seml']

    # Set Slurm job array options
    num_tasks = math.ceil(len(exp_array) / experiments_per_job)
    sbatch_options['array'] = f'0-{num_tasks - 1}'
    if max_simultaneous_jobs is not None:
        sbatch_options['array'] += f'%{max_simultaneous_jobs}'

    # Set Slurm output parameter
    if 'output' in sbatch_options:
        raise ConfigError(
            "Can't set sbatch `output` Parameter explicitly. SEML will do that for you."
        )
    elif output_dir_path == '/dev/null':
        output_file = output_dir_path
    else:
        output_file = f'{output_dir_path}/{name}_%A_%a.out'
        # Ensure that the output path exists
        Path(output_file).parent.mkdir(exist_ok=True)
    sbatch_options['output'] = output_file
    sbatch_options['job-name'] = name

    # Construct sbatch options string
    sbatch_options_str = create_slurm_options_string(sbatch_options, False)

    # Construct list with all experiment IDs
    expid_strings = f"{' '.join([str(exp['_id']) for exp in exp_array])}"

    with_sources = 'source_files' in seml_conf
    use_conda_env = seml_conf.get('conda_environment')
    working_dir = seml_conf.get('working_dir', '${{SLURM_SUBMIT_DIR}}')

    # Build arguments for the prepare_experiment script
    prepare_args = ''
    if with_sources:
        prepare_args += ' --stored-sources-dir $tmpdir'
    if logging.root.level <= logging.VERBOSE:
        prepare_args += ' --verbose'
    if unobserved:
        prepare_args += ' --unobserved'
    if debug_server:
        prepare_args += ' --debug-server'

    variables = {
        'sbatch_options': sbatch_options_str,
        'working_dir': working_dir,
        'use_conda_env': str(use_conda_env is not None).lower(),
        'conda_env': seml_conf['conda_environment'] if use_conda_env else '',
        'exp_ids': expid_strings,
        'with_sources': str(with_sources).lower(),
        'db_collection_name': collection.name,
        'prepare_args': prepare_args,
        'tmp_directory': SETTINGS.TMP_DIRECTORY,
        'experiments_per_job': experiments_per_job,
    }
    variables = {
        **variables,
        'setup_command': SETTINGS.SETUP_COMMAND.format(**variables),
        'end_command': SETTINGS.END_COMMAND.format(**variables),
    }
    # Construct Slurm script
    template = load_text_resource('templates/slurm/slurm_template.sh')
    script = template.format(**variables)

    # Dump the prepared script to a temporary file
    with NamedTemporaryFile('w', dir=SETTINGS.TMP_DIRECTORY) as f:
        f.write(script)
        f.flush()

        # Sbatch the script
        try:
            output = subprocess.run(
                f'sbatch {f.name}', shell=True, check=True, capture_output=True
            ).stdout
        except subprocess.CalledProcessError as e:
            logging.error(
                f"Could not start Slurm job via sbatch. Here's the sbatch error message:\n"
                f"{e.stderr.decode('utf-8')}"
            )
            exit(1)

    # Now we just update the mongodb. So, if we are in unobserved mode, we can stop here.
    if unobserved:
        return

    slurm_array_job_id = int(output.split(b' ')[-1])
    output_file = output_file.replace('%A', str(slurm_array_job_id))
    cluster_name = get_cluster_name()
    collection.update_many(
        {'_id': {'$in': [exp['_id'] for exp in exp_array]}},
        {
            '$set': {
                'status': States.PENDING[0],
                f'slurm.{slurm_options_id}.array_id': slurm_array_job_id,
                f'slurm.{slurm_options_id}.num_tasks': num_tasks,
                f'slurm.{slurm_options_id}.output_files_template': output_file,
                f'slurm.{slurm_options_id}.sbatch_options': sbatch_options,
                'execution.cluster': cluster_name,
            }
        },
    )
    return slurm_array_job_id


def start_srun_job(collection, exp, srun_options=None, seml_arguments=None):
    """Run a list of experiments as a job on the Slurm cluster.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    exp: dict
        Experiment to run.
    unobserved: bool
        Disable all Sacred observers (nothing written to MongoDB).
    srun_options: dict
        A dictionary that contains arguments for srun, e.g. {'mem': 8000} to limit the job's memory to 8,000 MB.
    seml_arguments: list
        A list that contains arguments for seml, e.g. ['--debug-server']

    Returns
    -------
    None
    """
    from seml.console import pause_live_widget

    with pause_live_widget():
        # Construct srun options string
        # srun will run 2 processes in parallel when ntasks is not specified. Probably because of hyperthreading.
        if 'ntasks' not in srun_options:
            srun_options['ntasks'] = 1
        srun_options_str = create_slurm_options_string(srun_options, True)

        # Set command args for job inside Slurm
        cmd_args = f"--local --sacred-id {exp['_id']} "
        cmd_args += ' '.join(seml_arguments)

        cmd = f'srun{srun_options_str} seml {collection.name} start {cmd_args}'
        try:
            subprocess.run(cmd, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            logging.error(
                f"Could not start Slurm job via srun. Here's the sbatch error message:\n"
                f"{e.stderr.decode('utf-8')}"
            )
            exit(1)


def start_local_job(
    collection,
    exp,
    unobserved=False,
    post_mortem=False,
    output_dir_path='.',
    output_to_console=False,
    debug_server=False,
):
    """Run an experiment locally.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    exp: dict
        Experiment to run.
    unobserved: bool
        Disable all Sacred observers (nothing written to MongoDB).
    post_mortem: bool
        Activate post-mortem debugging.
    output_dir_path: str
        Write the output to a file in `output_dir` given by the SEML config or in the current directory.
    output_to_console:
        Pipe all output (stdout and stderr) to the console.
    debug_server: bool
        Run job with a debug server.

    Returns
    -------
    True if job was executed successfully; False if it failed; None if job was not started because the database entry
    was not in the PENDING state.
    """
    from seml.console import pause_live_widget

    use_stored_sources = 'source_files' in exp['seml']

    interpreter, exe, config = get_command_from_exp(
        exp,
        collection.name,
        verbose=logging.root.level <= logging.VERBOSE,
        unobserved=unobserved,
        post_mortem=post_mortem,
        debug_server=debug_server,
    )
    cmd = get_shell_command(interpreter, exe, config)

    if not use_stored_sources:
        origin = Path().absolute()
        os.chdir(exp['seml']['working_dir'])

    success = True
    try:
        seml_config = exp['seml']
        slurm_config = exp['slurm']

        if use_stored_sources:
            temp_dir = os.path.join(SETTINGS.TMP_DIRECTORY, str(uuid.uuid4()))
            os.mkdir(temp_dir, mode=0o700)
            load_sources_from_db(exp, collection, to_directory=temp_dir)
            env = {'PYTHONPATH': f'{temp_dir}:$PYTHONPATH'}
            temp_exe = os.path.join(temp_dir, exe)
            # update the command to use the temp dir
            cmd = get_shell_command(interpreter, temp_exe, config, env=env)

        if output_dir_path:
            exp_name = get_exp_name(exp, collection.name)
            output_file = f"{output_dir_path}/{exp_name}_{exp['_id']}.out"
            if not unobserved:
                collection.update_one(
                    {'_id': exp['_id']}, {'$set': {'seml.output_file': output_file}}
                )
            if output_to_console:
                # redirect output to logfile AND output to console. See https://stackoverflow.com/a/34604684.
                # Alternatively, we could go with subprocess.Popen, but this could conflict with pdb.
                cmd = f'{cmd} 2>&1 | tee -a {output_file}'

        if seml_config.get('conda_environment') is not None:
            cmd = (
                f". $(conda info --base)/etc/profile.d/conda.sh "
                f"&& conda activate {seml_config['conda_environment']} "
                f"&& {cmd} "
                f"&& conda deactivate"
            )

        if not unobserved:
            execution = {'cluster': 'local'}
            if 'SLURM_JOBID' in os.environ:
                execution['array_id'] = os.environ['SLURM_JOBID']
                execution['task_id'] = 0
            collection.update_one(
                {'_id': exp['_id']},
                {'$set': {'execution': execution}},
            )

        logging.verbose(f'Running the following command:\n {cmd}')

        if output_dir_path:
            if output_to_console:
                # Let's pause the live widget so we can actually see the output.
                with pause_live_widget():
                    subprocess.run(cmd, shell=True, check=True)
            else:  # redirect output to logfile
                with open(output_file, 'w') as log_file:
                    subprocess.run(
                        cmd, shell=True, stderr=log_file, stdout=log_file, check=True
                    )
        else:
            with pause_live_widget():
                subprocess.run(cmd, shell=True, check=True)

    except subprocess.CalledProcessError:
        success = False
    except OSError:
        logging.error(f'Log file {output_file} could not be written.')
        # Since Sacred is never called in case of I/O error, we need to set the experiment state manually.
        if not unobserved:
            collection.update_one(
                filter={'_id': exp['_id']},
                update={'$set': {'status': States.FAILED[0]}},
            )
        success = False
    finally:
        if use_stored_sources and 'temp_dir' in locals():
            # clean up temp directory
            shutil.rmtree(temp_dir)
        if not use_stored_sources:
            os.chdir(origin)

    return success


def chunk_list(exps):
    """
    Divide experiments by batch id as these will be submitted jointly.
    This assumes constant Slurm settings per batch (which should be the case if MongoDB wasn't edited manually).

    Parameters
    ----------
    exps: list[dict]
        List of dictionaries containing the experiment settings as saved in the MongoDB

    Returns
    -------
    exp_chunks: list
    """
    from collections import defaultdict

    exp_chunks = defaultdict(list)
    for exp in exps:
        exp_chunks[exp['batch_id']].append(exp)
    return list(exp_chunks.values())


def prepare_staged_experiments(
    collection, filter_dict=None, num_exps=0, set_to_pending=True
):
    """
    Load experiments from the input MongoDB collection, and prepare them for running.
    If the filter_dict contains no status or ID, we filter the status by STAGED.
    If set_to_pending is True, we set their status to PENDING.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection with STAGED experiments.
    filter_dict: dict
        Optional dict with custom database filters.
    num_exps: int
        Only set <num_exps> experiments' state to PENDING. If 0, set all STAGED experiments to PENDING.
    set_to_pending: bool
        Whether to update the database entries to status PENDING.

    Returns
    -------
    The filtered list of database entries.
    """
    if filter_dict is None:
        filter_dict = {}

    query_dict = copy.deepcopy(filter_dict)
    if '_id' not in query_dict and 'status' not in query_dict:
        query_dict['status'] = {'$in': States.STAGED}

    experiments = list(collection.find(query_dict, limit=num_exps))

    if set_to_pending:
        update_dict = {'$set': {'status': States.PENDING[0]}}

        if num_exps > 0:
            # Set only those experiments to PENDING which will be run.
            collection.update_many(
                {'_id': {'$in': [e['_id'] for e in experiments]}}, update_dict
            )
        else:
            collection.update_many(query_dict, update_dict)

        nexps_set = len(experiments)
        logging.info(f'Setting {nexps_set} experiment{s_if(nexps_set)} to pending.')

    return experiments


def add_to_slurm_queue(
    collection,
    exps_list,
    unobserved=False,
    post_mortem=False,
    output_to_file=True,
    output_to_console=False,
    srun=False,
    debug_server=False,
):
    """
    Send the input list of experiments to the Slurm system for execution.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    exps_list: list of dicts
        The list of database entries corresponding to experiments to be executed.
    unobserved: bool
        Whether to suppress observation by Sacred observers.
    post_mortem: bool
        Activate post-mortem debugging.
    output_to_file: bool
        Whether to capture output in a logfile.
    output_to_console: bool
        Whether to capture output in the console. This is currently not supported for Slurm jobs and will raise an
        error if set to True.
    srun: bool
        Run jobs interactively via srun instead of using sbatch.
    debug_server: bool
        Run jobs with a debug server.

    Returns
    -------
    None
    """

    nexps = len(exps_list)
    exp_arrays = chunk_list(exps_list)
    narrays = 0
    array_ids = []

    for exp_array in exp_arrays:
        slurm_options = exp_array[0]['slurm']
        default_sbatch_options = slurm_options[0]['sbatch_options']
        job_name = get_exp_name(exp_array[0], collection.name)
        if srun:
            set_slurm_job_name(
                default_sbatch_options,
                job_name,
                exp_array[0],
                collection.name,
            )
            assert len(exp_array) == 1
            seml_arguments = []
            seml_arguments.append('--debug')
            if post_mortem:
                seml_arguments.append('--post-mortem')
            if output_to_console:
                seml_arguments.append('--output-to-console')
            if not output_to_file:
                seml_arguments.append('--no-file-output')
            if debug_server:
                seml_arguments.append('--debug-server')
            start_srun_job(
                collection,
                exp_array[0],
                srun_options=default_sbatch_options,
                seml_arguments=seml_arguments,
            )
            narrays += 1
        else:
            if output_to_file:
                output_dir_path = get_output_dir_path(exp_array[0])
            else:
                output_dir_path = '/dev/null'
            assert not post_mortem
            for slurm_options_id, slurm_option in enumerate(slurm_options):
                set_slurm_job_name(
                    slurm_option['sbatch_options'],
                    job_name,
                    exp_array[0],
                    collection.name,
                )
                array_id = start_sbatch_job(
                    collection,
                    exp_array,
                    slurm_options_id,
                    slurm_option['sbatch_options'],
                    unobserved,
                    name=job_name,
                    output_dir_path=output_dir_path,
                    max_simultaneous_jobs=slurm_option.get('max_simultaneous_jobs'),
                    experiments_per_job=slurm_option.get('experiments_per_job', 1),
                    debug_server=debug_server,
                )
                array_ids.append(array_id)
                narrays += 1
    logging.info(
        f'Started {nexps} experiment{s_if(nexps)} in '
        f'{narrays} Slurm job array{s_if(narrays)}: {", ".join(map(str, array_ids))}'
    )


def check_compute_node():
    if os.uname()[1] in SETTINGS.LOGIN_NODE_NAMES:
        raise ArgumentError(
            'Refusing to run a compute experiment on a login node. '
            'Please use Slurm or a compute node.'
        )


def start_local_worker(
    collection,
    num_exps=0,
    filter_dict=None,
    unobserved=False,
    post_mortem=False,
    steal_slurm=False,
    output_to_console=False,
    output_to_file=True,
    gpus=None,
    cpus=None,
    environment_variables=None,
    debug_server=False,
):
    """
    Start a local worker on the current machine that pulls PENDING experiments from the database and executes them.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    num_exps: int
        The maximum number of experiments run by this worker before terminating.
    filter_dict: dict
        Optional dict with custom database filters.
    unobserved: bool
        Whether to suppress observation by Sacred observers.
    post_mortem: bool
        Activate post-mortem debugging.
    steal_slurm: bool
        If True, the local worker will also execute jobs waiting for execution in Slurm.
    output_to_console: bool
        Whether to capture output in the console.
    output_to_file: bool
        Whether to capture output in a logfile.
    gpus: str
        Comma-separated list of GPU IDs to be used by this worker (e.g., "2,3"). Will be passed to CUDA_VISIBLE_DEVICES.
    cpus: int
        Number of CPU cores to be used by this worker. If None, use all cores.
    environment_variables: dict
        Optional dict of additional environment variables to be set.
    debug_server: bool
        Run jobs with a debug server.

    Returns
    -------
    None
    """
    from rich.progress import Progress

    from seml.console import pause_live_widget, prompt

    check_compute_node()

    if 'SLURM_JOBID' in os.environ:
        node_str = subprocess.run(
            'squeue -j ${SLURM_JOBID} -O nodelist:1000',
            shell=True,
            check=True,
            capture_output=True,
        ).stdout
        node_id = node_str.decode('utf-8').split('\n')[1].strip()
        logging.info(f'SLURM assigned me the node(s): {node_id}')

    if num_exps > 0:
        logging.info(
            f'Starting local worker thread that will run up to {num_exps} experiment{s_if(num_exps)}, '
            f'or until no pending experiments remain.'
        )
    else:
        logging.info(
            'Starting local worker thread that will run experiments until no pending experiments remain.'
        )
        num_exps = int(1e30)

    os.environ.update(get_environment_variables(gpus, cpus, environment_variables))

    num_exceptions = 0
    jobs_counter = 0

    exp_query = {}
    if not unobserved:
        exp_query['status'] = {'$in': States.PENDING}
    if not steal_slurm:
        exp_query['slurm'] = {'$elemMatch': {'array_id': {'$exists': False}}}

    exp_query.update(filter_dict)

    with pause_live_widget():
        with Progress(auto_refresh=False) as progress:
            task = progress.add_task('Running experiments...', total=None)
            while collection.count_documents(exp_query) > 0 and jobs_counter < num_exps:
                if unobserved:
                    exp = collection.find_one(exp_query)
                else:
                    exp = collection.find_one_and_update(
                        exp_query,
                        {
                            '$set': {
                                'status': States.RUNNING[0],
                                'execution.cluster': 'local',
                            }
                        },
                    )
                if exp is None:
                    continue
                if 'array_id' in exp.get('execution', {}):
                    # Clean up MongoDB entry
                    slurm_ids = {
                        'array_id': exp['execution']['array_id'],
                        'task_id': exp['execution']['task_id'],
                    }
                    reset_slurm_dict(exp)
                    collection.replace_one({'_id': exp['_id']}, exp, upsert=False)

                    # Cancel Slurm job; after cleaning up to prevent race conditions
                    if prompt(
                        f"SLURM is currently executing experiment {exp['_id']}, do you want to cancel the SLURM job?",
                        type=bool,
                    ):
                        cancel_experiment_by_id(
                            collection,
                            exp['_id'],
                            set_interrupted=False,
                            slurm_dict=slurm_ids,
                        )

                progress.console.print(
                    f"current id : {exp['_id']}, failed={num_exceptions}/{jobs_counter} experiments"
                )

                # Add newline if we need to avoid tqdm's output
                if (
                    debug_server
                    or output_to_console
                    or logging.root.level <= logging.VERBOSE
                ):
                    print(file=sys.stderr)

                if output_to_file:
                    output_dir_path = get_output_dir_path(exp)
                else:
                    output_dir_path = None
                try:
                    success = start_local_job(
                        collection=collection,
                        exp=exp,
                        unobserved=unobserved,
                        post_mortem=post_mortem,
                        output_dir_path=output_dir_path,
                        output_to_console=output_to_console,
                        debug_server=debug_server,
                    )
                    if success is False:
                        num_exceptions += 1
                except KeyboardInterrupt:
                    logging.info('Caught KeyboardInterrupt signal. Aborting.')
                    exit(1)
                jobs_counter += 1
                progress.advance(task)
                # tq.set_postfix(current_id=exp['_id'], failed=f"{num_exceptions}/{jobs_counter} experiments")


def start_experiments(
    db_collection_name,
    local,
    sacred_id,
    batch_id,
    filter_dict,
    num_exps,
    post_mortem,
    debug,
    debug_server,
    output_to_console,
    no_file_output,
    steal_slurm,
    no_worker,
    set_to_pending=True,
    worker_gpus=None,
    worker_cpus=None,
    worker_environment_vars=None,
):
    output_to_file = not no_file_output
    launch_worker = not no_worker

    if debug or debug_server:
        num_exps = 1
        unobserved = True
        post_mortem = True
        output_to_console = True
        srun = True
        logging.root.setLevel(logging.VERBOSE)
    else:
        unobserved = False
        srun = False

    if local:
        check_compute_node()

    if not local:
        local_kwargs = {
            '--no-worker': no_worker,
            '--steal-slurm': steal_slurm,
            '--worker-gpus': worker_gpus,
            '--worker-cpus': worker_cpus,
            '--worker-environment-vars': worker_environment_vars,
        }
        for key, val in local_kwargs.items():
            if val:
                raise ArgumentError(
                    f"The argument '{key}' only works in local mode, not in Slurm mode."
                )
    if not local and not srun:
        non_sbatch_kwargs = {
            '--post-mortem': post_mortem,
            '--output-to-console': output_to_console,
        }
        for key, val in non_sbatch_kwargs.items():
            if val:
                raise ArgumentError(
                    f"The argument '{key}' does not work in regular Slurm mode. "
                    "Remove the argument or use '--debug'."
                )

    if unobserved:
        set_to_pending = False

    filter_dict = build_filter_dict([], batch_id, filter_dict, sacred_id)

    collection = get_collection(db_collection_name)

    staged_experiments = prepare_staged_experiments(
        collection=collection,
        filter_dict=filter_dict,
        num_exps=num_exps,
        set_to_pending=set_to_pending and local,
    )

    if debug_server:
        use_stored_sources = 'source_files' in staged_experiments[0]['seml']
        if use_stored_sources:
            raise ArgumentError(
                'Cannot use a debug server with source code that is loaded from the MongoDB. '
                'Use the `--no-code-checkpoint` option when adding the experiment.'
            )

    if not local:
        add_to_slurm_queue(
            collection=collection,
            exps_list=staged_experiments,
            unobserved=unobserved,
            post_mortem=post_mortem,
            output_to_file=output_to_file,
            output_to_console=output_to_console,
            srun=srun,
            debug_server=debug_server,
        )
    elif launch_worker:
        start_local_worker(
            collection=collection,
            num_exps=num_exps,
            filter_dict=filter_dict,
            unobserved=unobserved,
            post_mortem=post_mortem,
            steal_slurm=steal_slurm,
            output_to_console=output_to_console,
            output_to_file=output_to_file,
            gpus=worker_gpus,
            cpus=worker_cpus,
            environment_variables=worker_environment_vars,
            debug_server=debug_server,
        )


def start_jupyter_job(
    sbatch_options: Optional[dict] = None,
    conda_env: Optional[str] = None,
    lab: bool = False,
):
    if lab:
        assert_package_installed(
            'jupyterlab',
            '`start-jupyter --lab` requires `jupyterlab` (e.g. `pip install jupyterlab`)',
        )
    else:
        assert_package_installed(
            'notebook',
            '`start-jupyter` requires `notebook` (e.g. `pip install notebook`)',
        )

    sbatch_options = sbatch_options if sbatch_options is not None else {}
    sbatch_options_merged = SETTINGS.SLURM_DEFAULT['sbatch_options']
    sbatch_options_merged.update(SETTINGS.SBATCH_OPTIONS_TEMPLATES.JUPYTER)
    sbatch_options_merged.update(sbatch_options)
    # Construct sbatch options string
    sbatch_options_str = create_slurm_options_string(sbatch_options_merged)

    template = load_text_resource('templates/slurm/jupyter_template.sh')

    script = template.format(
        sbatch_options=sbatch_options_str,
        use_conda_env=str(conda_env is not None).lower(),
        conda_env=conda_env,
        notebook_or_lab=' notebook' if not lab else '-lab',
    )

    path = os.path.join(SETTINGS.TMP_DIRECTORY, f'{uuid.uuid4()}.sh')
    with open(path, 'w') as f:
        f.write(script)

    try:
        output = subprocess.run(
            f'sbatch {path}', shell=True, check=True, capture_output=True
        ).stdout
    except subprocess.CalledProcessError as e:
        logging.error(
            f"Could not start Slurm job via sbatch. Here's the sbatch error message:\n"
            f"{e.stderr.decode('utf-8')}"
        )
        os.remove(path)
        exit(1)
    os.remove(path)

    slurm_array_job_id = int(output.split(b' ')[-1])
    logging.info(f'Queued Jupyter instance in Slurm job with ID {slurm_array_job_id}.')

    job_output = subprocess.run(
        f'scontrol show job {slurm_array_job_id} -o',
        shell=True,
        check=True,
        capture_output=True,
    ).stdout
    job_output_results = job_output.decode('utf-8').split(' ')
    job_info_dict = {
        x.split('=')[0]: x.split('=')[1] for x in job_output_results if '=' in x
    }
    log_file = job_info_dict['StdOut']

    logging.info(f"The job's log-file is '{log_file}'.")
    logging.info(
        'Waiting for start-up to fetch the machine and port of the Jupyter instance... '
        '(ctrl-C to cancel fetching)'
    )

    while job_info_dict['JobState'] in SlurmStates.PENDING:
        job_output = subprocess.run(
            f'scontrol show job {slurm_array_job_id} -o',
            shell=True,
            check=True,
            capture_output=True,
        ).stdout
        job_output_results = job_output.decode('utf-8').split(' ')
        job_info_dict = {
            x.split('=')[0]: x.split('=')[1] for x in job_output_results if '=' in x
        }
        time.sleep(1)
    if job_info_dict['JobState'] not in SlurmStates.RUNNING:
        logging.error(
            f"Slurm job failed. See log-file '{log_file}' for more information."
        )
        exit(1)

    logging.info('Slurm job is running. Jupyter instance is starting up...')

    # Obtain list of hostnames to addresses
    url_str, known_host = find_jupyter_host(log_file, True)
    if known_host is None:
        logging.error(
            f"Could not fetch the host and port of the Jupyter instance. Here's the raw output: \n"
            f'{url_str}'
        )
        exit(1)
    if not known_host:
        logging.warning('Host unknown to SLURM.')
    logging.info(f"Start-up completed. The Jupyter instance is running at '{url_str}'.")
    logging.info(f"To stop the job, run 'scancel {slurm_array_job_id}'.")


def get_experiment_to_prepare(
    collection: 'Collection',
    exp_id: int,
    unobserved: bool,
):
    """
    Retrieves the experiment the pending experiment with the given ID from the database and sets its state to RUNNING.

    Parameters
    ----------
    collection: pymongo.collection.Collection
        The MongoDB collection containing the experiments.
    exp_id: int
        The ID of the experiment to retrieve.
    unobserved: bool
        Whether to suppress observation by Sacred observers.

    Returns
    -------
    The experiment document if it was found and set to RUNNING, None otherwise.
    """
    if unobserved:
        # If the experiment has no observer, we just pull the configuration but never update the database.
        return collection.find_one({'_id': exp_id})
    # This returns the document as it was BEFORE the update. So we first have to check whether its state was
    # PENDING. This is to avoid race conditions, since find_one_and_update is an atomic operation.
    slurm_array_id, slurm_task_id = get_current_slurm_array_id()
    if slurm_array_id is not None and slurm_task_id is not None:
        # We're running in SLURM.
        # Check if the job executing is this one.
        cluster_name = get_cluster_name()
        job_filter = {
            '_id': exp_id,
            'execution.array_id': int(slurm_array_id),
            'execution.task_id': int(slurm_task_id),
            'execution.cluster': cluster_name,
        }
        # Either take the experiment if it is pending or if it is the one being executed.
        # The latter case is important for multi-node jobs.
        return collection.find_one(job_filter)
    # Steal slurm case
    return collection.find_one({'_id': exp_id})


def claim_experiment(db_collection_name: str, exp_ids: Sequence[int]):
    """
    Claim an experiment for execution by setting its state to RUNNING.

    Parameters
    ----------
    db_collection_name: str
        The name of the MongoDB collection containing the experiments.
    exp_ids: Sequence[int]
        The IDs of the experiments to claim.

    Exit Codes
    ----------
    0: Experiment claimed successfully
    3: Experiment not in the database

    Stdout
    -------
    The ID of the claimed experiment.
    """
    collection = get_collection(db_collection_name)
    array_id, task_id = get_current_slurm_array_id()
    if array_id is not None and task_id is not None:
        # We are running in slurm
        array_id, task_id = int(array_id), int(task_id)
        cluster_name = get_cluster_name()
        update = {
            'execution.array_id': array_id,
            'execution.task_id': task_id,
            'execution.cluster': cluster_name,
        }
        exp = collection.find_one_and_update(
            {'_id': {'$in': list(exp_ids)}, 'status': {'$in': States.PENDING}},
            {'$set': {'status': States.RUNNING[0], **update}},
            {'_id': 1, 'slurm': 1},
        )
        # Set slurm output file
        for s_conf in exp['slurm']:
            if s_conf['array_id'] == array_id:
                output_file = s_conf['output_files_template']
                output_file = output_file.replace('%a', str(task_id))
                collection.update_one(
                    {'_id': exp['_id']},
                    {'$set': {'execution.slurm_output_file': output_file}},
                )
    else:
        # Steal slurm
        exp = collection.find_one_and_update(
            {'_id': {'$in': list(exp_ids)}, 'status': {'$in': States.PENDING}},
            {'$set': {'status': States.RUNNING[0], 'execution.cluster': 'local'}},
            {'_id': 1},
        )
    if exp is None:
        exit(3)
    print(exp['_id'])
    exit(0)


def prepare_experiment(
    db_collection_name: str,
    exp_id: int,
    verbose: bool,
    unobserved: bool,
    post_mortem: bool,
    stored_sources_dir: Optional[str],
    debug_server: bool,
):
    """
    Prepare an experiment for execution by printing the command that should be executed.
    If stored_sources_dir is set, the source files are loaded from the database and stored in the directory.

    Parameters
    ----------
    db_collection_name: str
        The name of the MongoDB collection containing the experiments.
    exp_id: int
        The ID of the experiment to prepare.
    verbose: bool
        Whether to print the command verbosely.
    unobserved: bool
        Whether to suppress observation by Sacred observers.
    post_mortem: bool
        Activate post-mortem debugging.
    stored_sources_dir: str
        The directory where the source files are stored.
    debug_server: bool
        Run job with a debug server.

    Exit Codes
    ----------
    0: Preparation successful
    3: Experiment is not in the database
    4: Experiment is in the database but not in the PENDING state

    Returns
    -------
    None
    """
    from sacred.randomness import get_seed

    from seml.utils.multi_process import (
        is_local_main_process,
        is_main_process,
        is_running_in_multi_process,
    )

    # This process should only be executed once per node, so if we are not the main
    # process per node, we directly exit.
    if not is_local_main_process():
        exit(0)

    collection = get_collection(db_collection_name)
    exp = get_experiment_to_prepare(collection, exp_id, unobserved)

    if exp is None:
        # These exit codes will be handled in the bash script
        if collection.count_documents({'_id': exp_id}) == 0:
            exit(4)
        else:
            exit(3)

    if stored_sources_dir:
        os.makedirs(stored_sources_dir, exist_ok=True)
        if not os.listdir(stored_sources_dir):
            assert (
                'source_files' in exp['seml']
            ), '--stored-sources-dir is set but no source files are stored in the database.'
            load_sources_from_db(exp, collection, to_directory=stored_sources_dir)

    # The remaining part (updateing MongoDB & printing the python command) is only executed by the main process.
    if not is_main_process():
        exit(0)

    # If we run in a multi task environment, we want to make sure that the seed is fixed once and
    # all tasks start with the same seed. Otherwise, one could not reproduce the experiment as the
    # seed would change on the child nodes. It is up to the user to distribute seeds if needed.
    if is_running_in_multi_process():
        if SETTINGS.CONFIG_KEY_SEED not in exp['config']:
            exp['config'][SETTINGS.CONFIG_KEY_SEED] = get_seed()

    # Let's generate a output file
    output_dir = get_output_dir_path(exp)
    try:
        job_info = get_slurm_jobs(get_current_slurm_job_id())[0]
        name = job_info['JobName']
        array_id, task_id = get_current_slurm_array_id()
        name = f'{name}_{array_id}_{task_id}'
    except Exception:
        name = str(uuid.uuid4())
    output_file = f'{output_dir}/{name}_{exp["_id"]}.out'

    interpreter, exe, config = get_command_from_exp(
        exp,
        db_collection_name,
        verbose=verbose,
        unobserved=unobserved,
        post_mortem=post_mortem,
        debug_server=debug_server,
    )
    cmd = get_shell_command(interpreter, exe, config)
    cmd_unresolved = get_shell_command(
        *get_command_from_exp(
            exp,
            db_collection_name,
            verbose=verbose,
            unobserved=unobserved,
            post_mortem=post_mortem,
            debug_server=debug_server,
            unresolved=True,
        )
    )
    updates = {
        'seml.command': cmd,
        'seml.command_unresolved': cmd_unresolved,
        'seml.output_file': output_file,
    }

    if stored_sources_dir:
        temp_dir = stored_sources_dir
        # Store the temp dir for debugging purposes
        updates['seml.temp_dir'] = temp_dir
        cmd = get_shell_command(interpreter, os.path.join(temp_dir, exe), config)

    if not unobserved:
        collection.update_one({'_id': exp['_id']}, {'$set': updates})

    # Print the command to be ran.
    print(f'{cmd} > {output_file} 2>&1')
    # We exit with 0 to signal that the preparation was successful.
    exit(0)



import functools
import os
from contextlib import contextmanager
from typing import Sequence

import rich
import rich.box
import rich.progress
import rich.table
from rich.console import Console
from rich.padding import Padding
from rich.rule import Rule

from seml.utils.typer import prompt as typer_prompt

try:
    terminal_width = os.get_terminal_size().columns
except OSError:
    from seml.settings import SETTINGS

    terminal_width = SETTINGS.EXPERIMENT.TERMINAL_WIDTH

console = Console(
    width=terminal_width,
)

Table = functools.partial(
    rich.table.Table,
    collapse_padding=True,
    show_lines=False,
    show_edge=False,
    box=rich.box.SIMPLE,
    row_styles=['none', 'dim'],
    padding=(
        0,
        0,
    ),
    highlight=True,
)


@contextmanager
def pause_live_widget():
    prev_live = console._live
    if prev_live:
        prev_live.stop()
    console.clear_live()
    yield
    if prev_live:
        prev_live.start()


@functools.wraps(rich.progress.track)
def track(*args, **kwargs):
    """
    Wrapper for `rich.progress.track` that uses the global console instance and
    avoids creating empty ipython widgets in jupyter notebooks.

    Parameters
    ----------
    *args : Any
        Positional arguments to pass to `rich.progress.track`.
    **kwargs : Any
        Keyword arguments to pass to `rich.progress.track`.
    """
    # Directly return the sequence if the track is disabled. This avoids empty
    # ipywidgets in jupyter instances.
    if kwargs.get('disable', False):
        if len(args) == 0:
            yield from kwargs['sequence']
        else:
            yield from args[0]
        return

    if console not in kwargs:
        kwargs['console'] = console

    # Since there can only be one live instance at a time, we first need to stop
    # the previous one and then restart it after the new one is done.
    with pause_live_widget():
        yield from rich.progress.track(*args, **kwargs)


@functools.wraps(typer_prompt)
def prompt(*args, **kwargs):
    with pause_live_widget():
        return typer_prompt(*args, **kwargs)


def Heading(text: str):
    """
    Convenience function to create a seperator in the console.

    Parameters
    ----------
    text : str
        The text to display in the seperator.
    """
    return Padding(Rule(text, style='red'), pad=(1, 0, 0, 0))


def list_items(items: Sequence[str]):
    """
    Print a list of items in columns, using as many columns as possible
    while keeping the items readable. This will look similar to bash autcompletition suggestions.

    Parameters
    ----------
    items : Sequence[str]
        The items to print.
    """
    # Calculate available width
    available_width = console.width

    # Determine the number of columns based on available width
    max_len = max(len(s) for s in items)
    num_columns = max(1, available_width // (max_len + 1))

    # Create a table with the calculated number of columns
    table = rich.table.Table.grid(expand=True, pad_edge=True)
    for i in range(num_columns):
        table.add_column()

    # Add suggestions to the table, distributing them across columns
    for i in range(0, len(items), num_columns):
        table.add_row(*items[i : i + num_columns])

    # Print the table
    console.print(table)



import functools
import os
import sys
from typing import Callable, Optional, TypeVar, Union, overload

from typing_extensions import ParamSpec

_LOCAL_ID = 'SLURM_LOCALID'
_PROCESS_ID = 'SLURM_PROCID'
_PROCESS_COUNT = 'SLURM_NTASKS'


def process_id():
    return int(os.environ.get(_PROCESS_ID, 0))


def local_id():
    return int(os.environ.get(_LOCAL_ID, 0))


def process_count():
    return int(os.environ.get(_PROCESS_COUNT, 1))


def is_main_process():
    return process_id() == 0


def is_local_main_process():
    return local_id() == 0


def is_running_in_multi_process():
    return process_count() > 1


class ChildProcessSkip(Exception): ...


class MainProcessExecuteContext:
    def __enter__(self):
        if not is_main_process():
            sys.settrace(lambda *args, **keys: None)
            frame = sys._getframe(1)
            frame.f_trace = self.trace

    def trace(self, frame, event, arg):
        raise ChildProcessSkip()

    def __exit__(self, type, value, traceback):
        if type is None:
            return  # No exception
        if issubclass(type, ChildProcessSkip):
            return True  # Suppress special SkipWithBlock exception


P = ParamSpec('P')
R = TypeVar('R')


@overload
def only_on_main_process(func: Callable[P, R]) -> Callable[P, Optional[R]]: ...


@overload
def only_on_main_process(func: None = None) -> MainProcessExecuteContext: ...


def only_on_main_process(
    func: Optional[Callable[P, R]] = None,
) -> Union[Callable[P, Optional[R]], MainProcessExecuteContext]:
    if callable(func):

        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs):
            if is_main_process():
                return func(*args, **kwargs)
            return None

        return wrapper
    else:
        return MainProcessExecuteContext()



import logging
from typing import TYPE_CHECKING, Protocol

from seml.database import get_collection
from seml.settings import SETTINGS
from seml.utils import s_if, smaller_than_version_filter, utcnow
from seml.utils.slurm import get_cluster_name

if TYPE_CHECKING:
    from pymongo.collection import Collection

States = SETTINGS.STATES


def migrate_collection(db_collection_name: str, skip: bool, backup: bool):
    """
    Migrate the given collection to be compatible with the current SEML version.

    Parameters
    ----------
    db_collection_name : str
        The name of the collection to migrate.
    skip : bool
        Whether to skip the migration.
    backup : bool
        Whether to create a backup of the collection before migrating.
    """
    from seml.console import prompt

    skip = SETTINGS.MIGRATION.SKIP or skip
    yes = SETTINGS.MIGRATION.YES
    backup = SETTINGS.MIGRATION.BACKUP or backup
    backup_tmp = SETTINGS.MIGRATION.BACKUP_TMP

    if skip:
        return

    collection = get_collection(db_collection_name)
    migrations = [migration_cls(collection) for migration_cls in _MIGRATIONS]
    backed_up = False
    for migration in migrations:
        if not migration.requires_migration():
            continue

        if not migration.is_silent():
            logging.warning(
                f"The collection '{db_collection_name}' needs a {migration.name()} migration to work with newer SEML versions.\n"
                'If you wish to make backups, do not proceed and call `seml --migration-backup <collection> <command>`.\n'
                'To skip migration, use `seml --migration-skip <collection> <command>`.'
            )
            if not yes and not prompt('Do you want to proceed? (y/n)', type=bool):
                logging.error('Aborted migration.')
                exit(1)

        if not backed_up and backup and not migration.is_silent():
            backup_name = backup_tmp.format(
                collection=db_collection_name, time=utcnow().strftime('%Y%m%d_%H%M%S')
            )
            if backup_name in collection.database.list_collection_names():
                logging.error(f'Backup collection {backup_name} already exists.')
                exit(1)
            collection.aggregate([{'$match': {}}, {'$out': backup_name}])
            logging.info(
                f"Backed up collection '{db_collection_name}' to '{backup_name}'."
            )
            backed_up = True

        n_updated = migration.migrate()
        if not migration.is_silent():
            logging.info(
                f'Successfully migrated {n_updated} experiment{s_if(n_updated)}.'
            )


class Migration(Protocol):
    def __init__(self, collection: 'Collection'): ...
    def requires_migration(self) -> bool: ...
    def migrate(self) -> int: ...
    def name(self) -> str: ...
    def is_silent(self) -> bool: ...


class Migration04To05Slurm(Migration):
    collection: 'Collection'
    db_filter = {'slurm': {'$not': {'$type': 'array'}}}

    def __init__(self, collection: 'Collection'):
        self.collection = collection

    def is_silent(self):
        return False

    def requires_migration(self):
        return self.collection.count_documents(self.db_filter, limit=1) > 0

    def name(self):
        return 'SLURM config'

    def migrate(self):
        # Check if there are still experiments running
        # If so, we cannot migrate the SLURM configuration
        db_filter_running = {
            **self.db_filter,
            'status': {'$in': [*States.PENDING, *States.RUNNING]},
        }
        if self.collection.count_documents(db_filter_running) > 0:
            logging.error(
                'Cannot migrate SLURM configuration while there are still experiments running.\n'
                'Please wait until all experiments have finished.'
            )
            exit(1)

        # Move slurm array to execution field
        db_filter_executed = {**self.db_filter, 'slurm.array_id': {'$exists': True}}
        self.collection.update_many(
            db_filter_executed,
            {
                '$set': {
                    'execution': {
                        'cluster': get_cluster_name(),
                        'slurm.output_file': '$slurm.output_file',
                        'array_id': '$slurm.array_id',
                        'task_id': '$slurm.task_id',
                    }
                }
            },
        )
        # Convert slurm field to array
        n_updated = self.collection.update_many(
            self.db_filter,
            [{'$set': {'slurm': ['$slurm']}}],
        ).modified_count
        return n_updated


class Migration05Version(Migration):
    collection: 'Collection'
    db_filter = {
        '$or': [
            {f'seml.{SETTINGS.SEML_CONFIG_VALUE_VERSION}': {'$exists': False}},
            {
                f'seml.{SETTINGS.SEML_CONFIG_VALUE_VERSION}': {
                    '$not': {'$type': 'array'}
                }
            },
            smaller_than_version_filter((0, 5, 0)),
        ]
    }

    def __init__(self, collection: 'Collection'):
        self.collection = collection

    def is_silent(self):
        return True

    def requires_migration(self):
        return self.collection.count_documents(self.db_filter, limit=1) > 0

    def name(self):
        return 'Version'

    def migrate(self):
        n_updated = self.collection.update_many(
            self.db_filter,
            {'$set': {f'seml.{SETTINGS.SEML_CONFIG_VALUE_VERSION}': [0, 5, 0]}},
        ).modified_count
        return n_updated


_MIGRATIONS = [
    Migration04To05Slurm,
    Migration05Version,
]



from typing import Dict


def resolve_description(description: str, config: Dict) -> str:
    import uuid

    from omegaconf import OmegaConf

    # omegaconf can only resolve dicts that refers to its own values
    # so we add the description string to the config
    key = str(uuid.uuid4())
    config = OmegaConf.create(
        {key: description, **config}, flags={'allow_objects': True}
    )
    return OmegaConf.to_container(config, resolve=True)[key]



import logging
import os

from seml.database import get_mongo_client, get_mongodb_config
from seml.settings import SETTINGS
from seml.utils import warn_multiple_calls

__all__ = [
    'create_mongodb_observer',
    'create_slack_observer',
    'create_neptune_observer',
    'create_file_storage_observer',
    'add_to_file_storage_observer',
    'create_mattermost_observer',
]


@warn_multiple_calls(
    'Created {num_calls} MongoDB observers.\n'
    'This might not be intended.\n'
    'seml.experiment.Experiment creates one by default.\n'
    'Either disable the default observer (seml.experiment.Experiment(add_mongodb_observer=False)) \n'
    'or remove the explicit call to create_mongodb_observer.'
)
def create_mongodb_observer(collection, mongodb_config=None, overwrite=None):
    """Create a MongoDB observer for a Sacred experiment

    Parameters
    ----------
    collection: str
        Name of the collection in the database to write the results to.
    mongodb_config: dict
        Dictionary containing the connection details to the MongoDB. See get_mongodb_config().
    overwrite: int
        ID of an experiment to overwrite, e.g. a staged or failed experiment.

    Returns
    -------
    observer: MongoObserver
    """
    from sacred.observers import MongoObserver

    if mongodb_config is None:
        mongodb_config = get_mongodb_config()

    observer = MongoObserver(
        client=get_mongo_client(**mongodb_config),
        collection=collection,
        db_name=mongodb_config['db_name'],
        overwrite=overwrite,
    )
    return observer


def create_file_storage_observer(runs_folder_name, basedir=None, **kwargs):
    from sacred.observers import FileStorageObserver

    if basedir is None:
        basedir = SETTINGS.OBSERVERS.FILE.DEFAULT_BASE_DIR
        logging.info(
            f'Starting file observer in location {basedir}/{runs_folder_name}. To change the default base '
            f'directory, modify entry SETTINGS.OBSERVERS.FILE.DEFAULT_BASE_DIR in seml/settings.py.'
        )
    else:
        logging.info(
            f'Starting file observer in location {basedir}/{runs_folder_name}.'
        )
    observer = FileStorageObserver(f'{basedir}/{runs_folder_name}', kwargs)
    return observer


def add_to_file_storage_observer(file, experiment, delete_local_file=False):
    """

    Parameters
    ----------
    file: str
        Path to file to add to the file storage observer.
    experiment: sacred.experiment.Experiment
        The Sacred Experiment containing the FileStorageObserver.
    delete_local_file: bool, default: False
        If True, delete the local file after copying it to the FileStorageObserver.

    Returns
    -------
    None
    """
    has_file_observer = False
    for obs in experiment.current_run.observers:
        if 'FileStorageObserver' in str(type(obs)):
            obs.artifact_event(
                name=None,
                filename=file,
            )
            has_file_observer = True
    if not has_file_observer:
        logging.warning(
            "'add_to_file_storage_observer' was called but found no FileStorageObserver for the experiment."
        )
    if delete_local_file:
        os.remove(file)


def create_slack_observer(webhook=None):
    from sacred.observers import SlackObserver

    slack_obs = None
    if webhook is None:
        if 'OBSERVERS' in SETTINGS and 'SLACK' in SETTINGS.OBSERVERS:
            if 'WEBHOOK' in SETTINGS.OBSERVERS.SLACK:
                webhook = SETTINGS.OBSERVERS.SLACK.WEBHOOK
                slack_obs = SlackObserver(webhook)
    else:
        slack_obs = SlackObserver(webhook)

    if slack_obs is None:
        logging.warning('Failed to create Slack observer.')
    return slack_obs


def create_mattermost_observer(webhook=None, channel=None, **kwargs):
    """
    Create a Mattermost observer, which sends notifications via Mattermost.

    Parameters
    ----------
    webhook: str
        The webhook of the Mattermost instance. If you don't know this, ask your Mattermost administrator.
    channel: str
        The channel to send notifications to. This should usually be  @your.username (starting with '@').
    kwargs: dict
        Keyword arguments that are passed to the MattermostObserver. See MattermostObserver.__init__ for details.

    Returns
    -------
    The observer.

    """
    from seml.experiment.mattermost_observer import MattermostObserver

    if channel is None:
        if 'OBSERVERS' in SETTINGS and 'MATTERMOST' in SETTINGS.OBSERVERS:
            if channel is None and 'DEFAULT_CHANNEL' in SETTINGS.OBSERVERS.MATTERMOST:
                channel = SETTINGS.OBSERVERS.MATTERMOST.DEFAULT_CHANNEL
    if webhook is None:
        if 'OBSERVERS' in SETTINGS and 'MATTERMOST' in SETTINGS.OBSERVERS:
            if channel is None and 'DEFAULT_CHANNEL' in SETTINGS.OBSERVERS.MATTERMOST:
                channel = SETTINGS.OBSERVERS.MATTERMOST.DEFAULT_CHANNEL
            if 'WEBHOOK' in SETTINGS.OBSERVERS.MATTERMOST:
                webhook = SETTINGS.OBSERVERS.MATTERMOST.WEBHOOK
        else:
            raise ValueError('No webhook provided and none found in settings.py.')

    mattermost_observer = MattermostObserver(webhook, channel=channel, **kwargs)
    return mattermost_observer


def create_neptune_observer(
    project_name, api_token=None, source_extensions=['**/*.py', '**/*.yaml', '**/*.yml']
):
    try:
        from neptunecontrib.monitoring.sacred import NeptuneObserver
    except ImportError:
        logging.error(
            'Could not import neptunecontrib. Install via `pip install neptune-contrib`.'
        )

    if api_token is None:
        if 'OBSERVERS' in SETTINGS and 'NEPTUNE' in SETTINGS.OBSERVERS:
            if 'AUTH_TOKEN' in SETTINGS.OBSERVERS.NEPTUNE:
                api_token = SETTINGS.OBSERVERS.NEPTUNE.AUTH_TOKEN
                # Ignore example token setting
                if api_token == 'YOUR_AUTH_TOKEN':
                    api_token = None

    if api_token is None:
        logging.info(
            'No API token for Neptune provided. Trying to use environment variable NEPTUNE_API_TOKEN.'
        )
    neptune_obs = NeptuneObserver(
        api_token=api_token,
        project_name=project_name,
        source_extensions=source_extensions,
    )
    return neptune_obs



import os
import sys
from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_loader

AUTOCOMPLETING = bool(os.environ.get('_SEML_COMPLETE', False))


class PackageNotFoundError(Exception): ...


class FakeImportlibMetadata(Loader):
    def create_module(self, spec):
        return None

    def exec_module(self, module):
        module.PackageNotFoundError = PackageNotFoundError

        def version(package):
            return '0.0.0'

        module.version = version
        module.metadata = lambda name: {'version': '0.0.0'}


class ModuleHider(MetaPathFinder):
    def __init__(self, *hidden_modules: str, hide: bool = AUTOCOMPLETING) -> None:
        super().__init__()
        self.hidden_modules = set(hidden_modules)
        self.hide = hide

    def find_spec(self, fullname, path, target=None):
        if fullname in self.hidden_modules:
            # a special case for munch <3
            if fullname == 'importlib_metadata':
                return spec_from_loader(fullname, FakeImportlibMetadata())
            raise ImportError(f'No module named {fullname}')

    def __enter__(self):
        if self.hide:
            sys.meta_path.insert(0, self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.hide:
            sys.meta_path.remove(self)



import atexit
import logging
import time
from typing import Any, Dict

from seml.settings import SETTINGS
from seml.utils import assert_package_installed

States = SETTINGS.STATES


def retried_and_locked_ssh_port_forward(
    retries_max=SETTINGS.SSH_FORWARD.RETRIES_MAX,
    retries_delay=SETTINGS.SSH_FORWARD.RETRIES_DELAY,
    lock_file=SETTINGS.SSH_FORWARD.LOCK_FILE,
    lock_timeout=SETTINGS.SSH_FORWARD.LOCK_TIMEOUT,
    **ssh_config,
):
    """
    Attempt to establish an SSH tunnel with retries and a lock file to avoid parallel tunnel establishment.

    Parameters
    ----------
    retries_max: int
        Maximum number of retries to establish the tunnel.
    retries_delay: float
        Initial delay for exponential backoff.
    lock_file: str
        Path to the lock file.
    lock_timeout: int
        Timeout for acquiring the lock.
    ssh_config: dict
        Configuration for the SSH tunnel.

    Returns
    -------
    server: SSHTunnelForwarder
        The SSH tunnel server.
    """
    import random

    from filelock import FileLock, Timeout
    from sshtunnel import BaseSSHTunnelForwarderError, SSHTunnelForwarder

    delay = retries_delay
    error = None
    # disable SSH forward messages
    logging.getLogger('paramiko.transport').disabled = True
    for _ in range(retries_max):
        try:
            lock = FileLock(lock_file, mode=0o666, timeout=lock_timeout)
            with lock:
                server = SSHTunnelForwarder(**ssh_config)
                server.start()
                if not server.tunnel_is_up[server.local_bind_address]:
                    raise BaseSSHTunnelForwarderError()
                return server
        except Timeout as e:
            error = e
            logging.warn(f'Failed to aquire lock for ssh tunnel {lock_file}')
        except BaseSSHTunnelForwarderError as e:
            error = e
            logging.warn(f'Retry establishing ssh tunnel in {delay} s')
            # Jittered exponential retry
            time.sleep(delay)
            delay *= 2
            delay += random.uniform(0, 1)

    if error:
        logging.error(f'Failed to establish ssh tunnel: {error}')
        exit(1)


def _ssh_forward_process(pipe, ssh_config: Dict[str, Any]):
    """
    Establish an SSH tunnel in a separate process. The process periodically checks if the tunnel is still up and
    restarts it if it is not.

    Parameters
    ----------
    pipe: multiprocessing.communication.Connection
        Pipe to communicate with the main process.
    ssh_config: dict
        Configuration for the SSH tunnel.
    """
    server = retried_and_locked_ssh_port_forward(**ssh_config)
    # We need to bind to the same local addresses
    server._local_binds = server.local_bind_addresses
    pipe.send((server.local_bind_host, server.local_bind_port))
    while True:
        # check if we should end the process
        try:
            if pipe.poll(SETTINGS.SSH_FORWARD.HEALTH_CHECK_INTERVAL):
                if pipe.closed or pipe.recv() == 'stop':
                    server.stop()
                    break

            # Check for tunnel health
            server.check_tunnels()
            if not server.tunnel_is_up[server.local_bind_address]:
                logging.warning('SSH tunnel was closed unexpectedly. Restarting.')
                server.restart()
        except KeyboardInterrupt:
            server.stop()
            break
        except EOFError:
            server.stop()
            break
        except Exception as e:
            logging.error(f'Error in SSH tunnel health check:\n{e}')
            server.restart()
    pipe.close()


# We want to reuse the same multiprocessing context for all SSH tunnels
_mp_context = None


def get_forwarded_mongo_client(
    db_name, username, password, ssh_config: Dict[str, Any], **kwargs
):
    """
    Establish an SSH tunnel and return a forwarded MongoDB client.
    The SSH tunnel is established in a separate process to enable continuously checking for its health.

    Parameters
    ----------
    db_name: str
        Name of the database.
    username: str
        Username for the database.
    password: str
        Password for the database.
    ssh_config: dict
        Configuration for the SSH tunnel.
    kwargs: dict
        Additional arguments for the MongoDB client.

    Returns
    -------
    client: pymongo.MongoClient
        Forwarded MongoDB client.
    """
    import multiprocessing as mp

    import pymongo

    global _mp_context

    assert_package_installed(
        'sshtunnel',
        'Opening ssh tunnel requires `sshtunnel` (e.g. `pip install sshtunnel`)',
    )
    assert_package_installed(
        'filelock',
        'Opening ssh tunnel requires `filelock` (e.g. `pip install filelock`)',
    )

    if _mp_context is None:
        _mp_context = mp.get_context('forkserver')
    main_pipe, forward_pipe = _mp_context.Pipe(True)
    proc = _mp_context.Process(
        target=_ssh_forward_process, args=(forward_pipe, ssh_config)
    )
    proc.start()

    def try_close():
        try:
            if not main_pipe.closed:
                main_pipe.send('stop')
                main_pipe.close()
        finally:
            pass

    class ForwardedMongoClient(pymongo.MongoClient):
        def __del__(self):
            try_close()

    # Send stop if we exit the program
    atexit.register(try_close)

    # Compute the maximum time we should wait
    retries_max = ssh_config.get('retries_max', SETTINGS.SSH_FORWARD.RETRIES_MAX)
    retries_delay = ssh_config.get('retries_delay', SETTINGS.SSH_FORWARD.RETRIES_DELAY)
    max_delay = 2 ** (retries_max + 1) * retries_delay

    # check if the forward process has been established correctly
    if main_pipe.poll(max_delay):
        host, port = main_pipe.recv()

        client = ForwardedMongoClient(
            host,
            int(port),
            username=username,
            password=password,
            authSource=db_name,
            **kwargs,
        )
        return client
    else:
        logging.error('Failed to establish SSH tunnel.')
        exit(1)



import array
import fcntl
import socket
import struct
import sys
from contextlib import closing


def get_network_interfaces():
    # From https://code.activestate.com/recipes/439093/#c1
    is_64bits = sys.maxsize > 2**32
    struct_size = 40 if is_64bits else 32
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    max_possible = 8  # initial value
    while True:
        _bytes = max_possible * struct_size
        names = array.array('B')
        for i in range(0, _bytes):
            names.append(0)
        outbytes = struct.unpack(
            'iL',
            fcntl.ioctl(
                s.fileno(),
                0x8912,  # SIOCGIFCONF
                struct.pack('iL', _bytes, names.buffer_info()[0]),
            ),
        )[0]
        if outbytes == _bytes:
            max_possible *= 2
        else:
            break
    namestr = names.tobytes()
    ifaces = {}
    for i in range(0, outbytes, struct_size):
        iface_name = bytes.decode(namestr[i : i + 16]).split('\0', 1)[0]
        iface_addr = socket.inet_ntoa(namestr[i + 20 : i + 24])
        ifaces[iface_name] = iface_addr

    return ifaces


def find_free_port():
    ifaces = get_network_interfaces()
    if 'lo' in ifaces:
        del ifaces['lo']
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind((list(ifaces.values())[0], 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        ip_address, port = s.getsockname()[:2]
        return ip_address, port



class InputError(SystemExit):
    """Parent class for input errors that don't print a stack trace."""

    pass


class ConfigError(InputError):
    """Raised when the something is wrong in the config"""

    def __init__(self, message='The config file contains an error.'):
        super().__init__(f'CONFIG ERROR: {message}')


class ExecutableError(InputError):
    """Raised when the something is wrong with the Python executable"""

    def __init__(self, message='The Python executable has a problem.'):
        super().__init__(f'EXECUTABLE ERROR: {message}')


class MongoDBError(InputError):
    """Raised when the something is wrong with the MongoDB"""

    def __init__(self, message='The MongoDB or its config has a problem.'):
        super().__init__(f'MONGODB ERROR: {message}')


class ArgumentError(InputError):
    """Raised when the something is wrong with the parsed arguments"""

    def __init__(self, message='The parsed arguments contain an error.'):
        super().__init__(f'ARGUMENT ERROR: {message}')
